""" Module to decouple cg code from Housekeeper code """

import logging
import os
from datetime import datetime
from pathlib import Path

from housekeeper.include import checksum as hk_checksum
from housekeeper.include import include_version
from housekeeper.store.database import create_all_tables, drop_all_tables, initialize_database
from housekeeper.store.models import Archive, Bundle, File, Tag, Version
from housekeeper.store.store import Store
from sqlalchemy.orm import Query

from cg.constants import SequencingFileTag
from cg.exc import (
    HousekeeperArchiveMissingError,
    HousekeeperBundleVersionMissingError,
    HousekeeperFileMissingError,
)

LOG = logging.getLogger(__name__)


class HousekeeperAPI:
    """API to decouple cg code from Housekeeper."""

    def __init__(self, config: dict) -> None:
        initialize_database(config["housekeeper"]["database"])
        self._store = Store(config["housekeeper"]["root"])
        self.root_dir: str = config["housekeeper"]["root"]

    def new_bundle(self, name: str, created_at: datetime = None) -> Bundle:
        """Create a new file bundle."""
        return self._store.new_bundle(name, created_at)

    def add_bundle(self, bundle_data) -> tuple[Bundle, Version]:
        """Build a new bundle version of files."""
        return self._store.add_bundle(bundle_data)

    def bundle(self, name: str) -> Bundle:
        """Fetch a bundle."""
        return self._store.get_bundle_by_name(bundle_name=name)

    def bundles(self) -> list[Bundle]:
        """Fetch bundles."""
        return self._store.bundles()

    def create_new_bundle_and_version(self, name: str) -> Bundle:
        """Create new bundle with version."""
        new_bundle: Bundle = self.new_bundle(name=name)
        self.add_commit(new_bundle)
        new_version: Version = self.new_version(created_at=new_bundle.created_at)
        new_bundle.versions.append(new_version)
        self.commit()
        LOG.info(f"New bundle created with name {new_bundle.name}")
        return new_bundle

    def set_to_archive(self, file: File, value: bool) -> None:
        """Sets the 'to_archive' field of a file."""
        file.to_archive = value
        self.commit()

    def new_file(
        self, path: str, checksum: str = None, to_archive: bool = False, tags: list = None
    ) -> File:
        """Create a new file."""
        if tags is None:
            tags = []
        return self._store.new_file(path, checksum, to_archive, tags)

    def get_file(self, file_id: int) -> File | None:
        """Get a file based on file id."""
        LOG.info(f"Return file: {file_id}")
        file_obj: File = self._store.get_file_by_id(file_id=file_id)
        if not file_obj:
            LOG.info("file not found")
            return None
        return file_obj

    def delete_file(self, file_id: int) -> File | None:
        """Delete a file both from database and disk (if included)."""
        file_obj: File = self.get_file(file_id)
        if not file_obj:
            LOG.info(f"Could not find file {file_id}")
            return

        if file_obj.is_included and Path(file_obj.full_path).exists():
            LOG.info(f"Deleting file {file_obj.full_path} from disc")
            Path(file_obj.full_path).unlink()

        LOG.info(f"Deleting file {file_id} from housekeeper")
        self._store.session.delete(file_obj)
        self.commit()

        return file_obj

    def add_file(
        self, path: str, version_obj: Version, tags: list, to_archive: bool = False
    ) -> File:
        """Add a file to the database."""
        if isinstance(tags, str):
            tags: list[str] = [tags]
        for tag_name in tags:
            if not self.get_tag(tag_name):
                self.add_tag(tag_name)

        new_file: File = self.new_file(
            path=str(Path(path).absolute()),
            to_archive=to_archive,
            tags=[self.get_tag(tag_name) for tag_name in tags],
        )

        new_file.version = version_obj
        self._store.session.add(new_file)
        return new_file

    def files(
        self,
        *,
        bundle: str = None,
        tags: set[str] = None,
        version: int | None = None,
        path: str = None,
    ) -> Query:
        """Fetch files."""
        return self._store.get_files(
            bundle_name=bundle, tag_names=tags, version_id=version, file_path=path
        )

    def get_file_insensitive_path(self, path: Path) -> File | None:
        """Returns a file in Housekeeper given any kind of path (absolute or relative)."""
        file: File = self.files(path=path.as_posix()).first()
        if not file:
            if path.is_absolute():
                file = self.files(path=str(path).replace(f"{self.root_dir}/", "")).first()
            else:
                file = self.files(path=f"{self.root_dir}/{path}").first()
        return file

    @staticmethod
    def get_files_from_version(version: Version, tags: set[str]) -> list[File]:
        """Return a list of files associated with the given version and tags."""
        LOG.debug(f"Getting files from version with tags {tags}")
        files: list[File] = []
        for file in list(version.files):
            file_tags = {tag.name for tag in file.tags}
            if tags.issubset(file_tags):
                LOG.debug(f"Found file {file}")
                files.append(file)
        if not files:
            LOG.warning(f"Could not find any files matching the tags {tags}")
        return files

    @staticmethod
    def get_file_from_version(version: Version, tags: set[str]) -> File | None:
        """Return the first file matching the given tags."""
        files: list[File] = HousekeeperAPI.get_files_from_version(version=version, tags=tags)
        return next((file for file in files), None)

    @staticmethod
    def get_latest_file_from_version(version: Version, tags: set[str]) -> File | None:
        """Return the latest file from Housekeeper given its version and tags."""
        files: list[File] = HousekeeperAPI.get_files_from_version(version=version, tags=tags)
        return sorted(files, key=lambda file_obj: file_obj.id)[-1] if files else None

    def rollback(self):
        """Wrap method in Housekeeper Store."""
        return self._store.session.rollback()

    def get_files(self, bundle: str, tags: list | None = None, version: int | None = None) -> Query:
        """Get all the files in housekeeper, optionally filtered by bundle and/or tags and/or
        version.
        """
        return self._store.get_files(bundle_name=bundle, tag_names=tags, version_id=version)

    def get_latest_file(
        self, bundle: str, tags: list | None = None, version: int | None = None
    ) -> File | None:
        """Return latest file from Housekeeper, filtered by bundle and/or tags and/or version."""
        files: Query = self._store.get_files(bundle_name=bundle, tag_names=tags, version_id=version)
        return files.order_by(File.id.desc()).first()

    def get_latest_file_strict(
        self, bundle: str, tags: list | None = None, version: int | None = None
    ) -> File:
        """
        Return latest file from Housekeeper, filtered by bundle and/or tags and/or version.
        Raises:
            HousekeeperFileMissingError if no file is found.
        """
        files: Query = self._store.get_files(bundle_name=bundle, tag_names=tags, version_id=version)
        latest_file: File | None = files.order_by(File.id.desc()).first()
        if not latest_file:
            raise HousekeeperFileMissingError(
                f"No file matching bundle {bundle}, version {version} with tags {tags}"
            )
        else:
            return latest_file

    def check_bundle_files(
        self,
        bundle_name: str,
        file_paths: list[Path],
        last_version: Version,
        tags: list | None = None,
    ) -> list[Path]:
        """Checks if any of the files in the provided list are already added to the provided
        bundle. Returns a list of files that have not been added."""
        for file in self.get_files(bundle=bundle_name, tags=tags, version=last_version.id):
            if Path(file.path) in file_paths:
                file_paths.remove(Path(file.path))
                LOG.info(
                    f"Path {file.path} is already linked to bundle {bundle_name} in housekeeper"
                )
        return file_paths

    @staticmethod
    def get_included_path(root_dir: Path, version_obj: Version, file_obj: File) -> Path:
        """Generate the path to a file that should be included.
        If the version dir does not exist, create a new version dir in root dir.
        """
        version_root_dir: Path = Path(root_dir, version_obj.relative_root_dir)
        version_root_dir.mkdir(parents=True, exist_ok=True)
        LOG.info(f"Created new bundle version dir: {version_root_dir}")
        return Path(version_root_dir, Path(file_obj.path).name)

    def include_file(self, file_obj: File, version_obj: Version) -> File:
        """Call the include version function to import related assets."""
        global_root_dir: Path = Path(self.get_root_dir())

        new_path: Path = self.get_included_path(
            root_dir=global_root_dir, version_obj=version_obj, file_obj=file_obj
        )
        if file_obj.to_archive:
            # calculate sha1 checksum if file is to be archived
            file_obj.checksum = HousekeeperAPI.checksum(file_obj.path)
        if new_path.exists():
            LOG.warning(
                f"Another file with identical included file path: {new_path} already exist. Skip linking of: {file_obj.path}"
            )
            file_obj.path = str(new_path).replace(f"{global_root_dir}/", "", 1)
            return file_obj
        # hardlink file to the internal structure
        os.link(file_obj.path, new_path)
        LOG.info(f"Linked file: {file_obj.path} -> {new_path}")
        file_obj.path = str(new_path).replace(f"{global_root_dir}/", "", 1)
        return file_obj

    def new_version(self, created_at: datetime, expires_at: datetime = None) -> Version:
        """Create a new bundle version."""
        return self._store.new_version(created_at, expires_at)

    def get_version_by_id(self, version_id: int) -> Version:
        """Return a version given its database ID."""
        return self._store.get_version_by_id(version_id=version_id)

    def last_version(self, bundle: str) -> Version:
        """Gets the latest version of a bundle."""
        LOG.debug(f"Fetch latest version from bundle {bundle}")
        return (
            self._store._get_query(table=Version)
            .join(Version.bundle)
            .filter(Bundle.name == bundle)
            .order_by(Version.created_at.desc())
            .first()
        )

    def get_non_archived_spring_files(
        self, tags: list[str] | None = None, limit: int = None
    ) -> list[File]:
        """Return all spring files which are not marked as archived in Housekeeper."""
        return self._store.get_non_archived_files(
            tag_names=tags + [SequencingFileTag.SPRING] if tags else [SequencingFileTag.SPRING],
            limit=limit,
        )

    def get_latest_bundle_version(self, bundle_name: str) -> Version | None:
        """Get the latest version of a Housekeeper bundle."""
        last_version: Version = self.last_version(bundle_name)
        if not last_version:
            LOG.debug(f"No bundle found for {bundle_name} in Housekeeper")
            return None
        LOG.debug(f"Found Housekeeper version object for {bundle_name}: {repr(last_version)}")
        return last_version

    def get_or_create_version(self, bundle_name: str) -> Version:
        """Returns the latest version of a bundle if it exists. If not creates a bundle and
        returns its version."""
        last_version: Version = self.last_version(bundle=bundle_name)
        if not last_version:
            LOG.info(f"Creating bundle for sample {bundle_name} in housekeeper")
            bundle_result: tuple[Bundle, Version] = self.add_bundle(
                bundle_data={
                    "name": bundle_name,
                    "created_at": datetime.now(),
                    "expires_at": None,
                    "files": [],
                }
            )
            last_version: Version = bundle_result[1]
        return last_version

    def new_tag(self, name: str, category: str = None):
        """Create a new tag."""
        return self._store.new_tag(name, category)

    def add_tag(self, name: str, category: str = None) -> Tag:
        """Add a tag to the database."""
        tag_obj = self._store.new_tag(name, category)
        self.add_commit(tag_obj)
        return tag_obj

    def get_tag(self, name: str) -> Tag:
        """Fetch a tag."""
        return self._store.get_tag(name)

    @staticmethod
    def get_tag_names_from_file(file: File) -> list[str]:
        """Fetch tag names for a file."""
        return [tag.name for tag in file.tags]

    def include(self, version_obj: Version):
        """Call the include version function to import related assets."""
        include_version(self.get_root_dir(), version_obj)
        version_obj.included_at = datetime.now()

    def add_commit(self, obj):
        """Wrap method in Housekeeper Store."""
        self._store.session.add(obj)
        return self._store.session.commit()

    def commit(self):
        """Wrap method in Housekeeper Store."""
        return self._store.session.commit()

    def get_root_dir(self) -> str:
        """Returns the root dir of Housekeeper."""
        return self.root_dir

    @staticmethod
    def checksum(path) -> str:
        """Calculate the checksum."""
        return hk_checksum(path)

    def initialise_db(self):
        """Create all tables in the store."""
        create_all_tables()

    def destroy_db(self):
        """Drop all tables in the store."""
        drop_all_tables()

    def add_and_include_file_to_latest_version(
        self, bundle_name: str, file: Path, tags: list
    ) -> None:
        """Adds and includes a file in the latest version of a bundle."""
        version: Version = self.get_or_create_version(bundle_name)
        hk_file: File = self.add_file(version_obj=version, tags=tags, path=str(file.absolute()))
        self.include_file(version_obj=version, file_obj=hk_file)
        self.commit()

    def include_files_to_latest_version(self, bundle_name: str) -> None:
        """Include all files in the latest version on a bundle."""
        bundle_version: Version = self.get_latest_bundle_version(bundle_name=bundle_name)
        if not bundle_version:
            return None
        if bundle_version.included_at:
            LOG.info(
                f"Bundle: {bundle_name}, version: {bundle_version} already included at {bundle_version.included_at}"
            )
            return
        for hk_file in bundle_version.files:
            if not hk_file.is_included:
                try:
                    self.include_file(version_obj=bundle_version, file_obj=hk_file)
                except FileExistsError as error:
                    LOG.error(error)
                continue
            LOG.warning(
                f"File is already included in Housekeeper for bundle: {bundle_name}, version: {bundle_version}"
            )
        bundle_version.included_at = datetime.now()
        self.commit()

    def get_file_from_latest_version(self, bundle_name: str, tags: set[str]) -> File | None:
        """Return a file in the latest version of a bundle."""
        version: Version = self.last_version(bundle=bundle_name)
        if not version:
            LOG.warning(f"Bundle: {bundle_name} not found in Housekeeper")
            raise HousekeeperBundleVersionMissingError
        return self.files(version=version.id, tags=tags).first()

    def get_files_from_latest_version(
        self, bundle_name: str, tags: list[str] | None = None
    ) -> list[File] | None:
        """Return files in the latest version of a bundle.

        Raises HousekeeperBundleVersionMissingError:
        - When no version was found for the given bundle
        """
        version: Version = self.last_version(bundle=bundle_name)
        if not version:
            LOG.warning(f"Bundle: {bundle_name} not found in Housekeeper")
            raise HousekeeperBundleVersionMissingError
        return self.files(version=version.id, tags=tags).all()

    def get_non_archived_files_for_bundle(
        self, bundle_name: str, tags: list | None = None
    ) -> list[File]:
        """Returns all non-archived_files from a given bundle, tagged with the given tags"""
        return self._store.get_non_archived_files_for_bundle(
            bundle_name=bundle_name, tags=tags or []
        )

    def get_archived_files_for_bundle(
        self, bundle_name: str, tags: list | None = None
    ) -> list[File]:
        """Returns all archived_files from a given bundle, tagged with the given tags"""
        LOG.debug(f"Getting archived files for bundle {bundle_name}")
        return self._store.get_archived_files_for_bundle(bundle_name=bundle_name, tags=tags or [])

    def get_archived_files_not_being_retrieved_for_bundle(
        self, bundle_name: str, tags: list | None = None
    ) -> list[File]:
        """Returns archived files from a bundle, excluding files currently being retrieved."""
        LOG.debug(f"Getting archived files which are not being retrieved for bundle {bundle_name}.")
        return self._store.get_archived_files_for_bundle_excluding_ongoing_retrievals(
            bundle_name=bundle_name, tags=tags or []
        )

    def add_archives(self, files: list[File], archive_task_id: int) -> None:
        """Creates an archive object for the given files, and adds the archive task id to them."""
        for file in files:
            file_id: int = file.id
            LOG.info(
                f"Adding an archive to file {file_id} with archiving task id {archive_task_id}"
            )
            archive: Archive = self._store.create_archive(
                file_id=file_id, archiving_task_id=archive_task_id
            )
            self._store.session.add(archive)
        self.commit()

    def set_archive_retrieved_at(self, file_id: int, retrieval_task_id: int):
        """Sets the retrieved_at value for an Archive entry. Raises a ValueError if the given retrieval task id
        is not found in Housekeeper."""
        archive: Archive = self._store.get_file_by_id(file_id).archive
        if archive.retrieval_task_id != retrieval_task_id:
            raise ValueError(
                f"Retrieval task id did not match database entry. Given task id was {retrieval_task_id}, "
                f"while retrieval task id in Housekeeper is {archive.retrieval_task_id}."
            )
        self._store.update_retrieval_time_stamp(archive=archive)
        self.commit()

    def set_archive_archived_at(self, file_id: int, archiving_task_id: int):
        """Sets the archived_at value for an Archive entry. Raises a ValueError if the given archiving task id
        is not found in Housekeeper."""
        archive: Archive = self._store.get_file_by_id(file_id).archive
        if archive.archiving_task_id != archiving_task_id:
            raise ValueError(
                f"Archiving task id did not match database entry. Given task id was {archiving_task_id}, "
                f"while archiving task id in Housekeeper is {archive.archiving_task_id}."
            )
        self._store.update_archiving_time_stamp(archive=archive)

    def set_archive_retrieval_task_id(self, file_id: int, retrieval_task_id: int) -> None:
        """Sets the retrieval_task_id for an Archive entry. Raises a ValueError if the given retrieval task id
        is not found in Housekeeper."""
        archive: Archive = self._store.get_file_by_id(file_id).archive
        if not archive:
            raise ValueError(f"No Archive entry found for file with id {file_id}.")
        self._store.update_retrieval_task_id(archive=archive, retrieval_task_id=retrieval_task_id)
        self.commit()

    def get_sample_sheets_from_latest_version(self, flow_cell_id: str) -> list[File]:
        """Returns the files tagged with 'samplesheet' for the given bundle."""
        try:
            sample_sheet_files: list[File] = self.get_files_from_latest_version(
                bundle_name=flow_cell_id, tags=[flow_cell_id, SequencingFileTag.SAMPLE_SHEET]
            )
        except HousekeeperBundleVersionMissingError:
            sample_sheet_files = []
        return sample_sheet_files

    def get_sample_sheet_path(self, flow_cell_id: str) -> Path:
        """Returns the sample sheet path for the flow cell."""
        sample_sheet_files: list[File] = self.get_sample_sheets_from_latest_version(flow_cell_id)
        if not sample_sheet_files:
            LOG.error(f"Sample sheet file for flowcell {flow_cell_id} not found in Housekeeper!")
            raise HousekeeperFileMissingError
        return Path(sample_sheet_files[0].full_path)

    def file_exists_in_latest_version_for_bundle(self, file_path: Path, bundle_name: str) -> bool:
        """Check if a file exists in the latest version for bundle."""
        latest_version: Version = self.get_latest_bundle_version(bundle_name)
        return any(
            file_path.name == Path(bundle_file.path).name for bundle_file in latest_version.files
        )

    def add_file_to_bundle_if_non_existent(
        self, file_path: Path, bundle_name: str, tag_names: list[str]
    ) -> None:
        """Add file to Housekeeper if it has not already been added."""
        if not file_path.exists():
            LOG.warning(f"File does not exist: {file_path}")
            return

        if not self.file_exists_in_latest_version_for_bundle(
            file_path=file_path, bundle_name=bundle_name
        ):
            self.add_and_include_file_to_latest_version(
                bundle_name=bundle_name,
                file=file_path,
                tags=tag_names,
            )
            LOG.info(f"File added to Housekeeper bundle {bundle_name}")
        else:
            LOG.info(f"Bundle {bundle_name} already has a file with the same name as {file_path}")

    def add_tags_if_non_existent(self, tag_names: list[str]) -> None:
        """Ensure that tags exist in Housekeeper."""
        for tag_name in tag_names:
            if self.get_tag(name=tag_name) is None:
                self.add_tag(name=tag_name)

    def add_bundle_and_version_if_non_existent(self, bundle_name: str) -> None:
        """Add bundle if it does not exist."""
        if not self.bundle(name=bundle_name):
            self.create_new_bundle_and_version(name=bundle_name)
        else:
            LOG.debug(f"Bundle with name {bundle_name} already exists")

    def create_bundle_and_add_file_with_tags(
        self,
        bundle_name: str,
        file_path: Path,
        tags: list[str],
    ) -> None:
        """Add a file path with tags to a bundle and version in Housekeeper."""
        self.add_bundle_and_version_if_non_existent(bundle_name)
        self.add_tags_if_non_existent([bundle_name])
        self.add_file_to_bundle_if_non_existent(
            file_path=file_path,
            bundle_name=bundle_name,
            tag_names=tags,
        )

    def get_archive_entries(
        self, archival_task_id: int = None, retrieval_task_id: int = None
    ) -> list[Archive]:
        """Returns all archives matching the provided task ids. If no task ids are provided, all archive entries are
        returned. If only an archival_task_id is provided, filtering is only done on that parameter and vice versa
        with retrieval_task_id.
        """
        return self._store.get_archives(
            archival_task_id=archival_task_id, retrieval_task_id=retrieval_task_id
        )

    def set_archived_at(self, archival_task_id: int) -> None:
        """Sets archived_at to the current time for archive entries with matching archival task id.
        Raises:
                HousekeeperArchiveMissingError if no Archive entries match the given retrieval task id.
        """
        archive_entries: list[Archive] = self.get_archive_entries(archival_task_id=archival_task_id)
        if not archive_entries:
            raise HousekeeperArchiveMissingError(
                f"Could not find any archives with archival_task_id {archival_task_id}"
            )
        for archive in archive_entries:
            self.set_archive_archived_at(
                archiving_task_id=archival_task_id, file_id=archive.file_id
            )
        self.commit()

    def set_retrieved_at(self, retrieval_task_id: int) -> None:
        """Sets retrieved_at to the current time for archive entries with matching archival task id.
        Raises:
                HousekeeperArchiveMissingError if no Archive entries match the given retrieval task id.
        """
        archive_entries: list[Archive] = self.get_archive_entries(
            retrieval_task_id=retrieval_task_id
        )
        if not archive_entries:
            raise HousekeeperArchiveMissingError(
                f"Could not find any archives with retrieval_task_id {retrieval_task_id}"
            )
        for archive in archive_entries:
            self.set_archive_retrieved_at(
                retrieval_task_id=retrieval_task_id, file_id=archive.file_id
            )
        self.commit()

    def get_ongoing_archivals(self) -> list[Archive]:
        return self._store.get_ongoing_archivals()

    def get_ongoing_retrievals(self) -> list[Archive]:
        return self._store.get_ongoing_retrievals()

    def delete_archives(self, archival_task_id: int):
        archives_to_delete: list[Archive] = self.get_archive_entries(
            archival_task_id=archival_task_id
        )
        for archive in archives_to_delete:
            self._store.session.delete(archive)
        self._store.session.commit()

    def update_archive_retrieved_at(
        self, old_retrieval_job_id: int, new_retrieval_job_id: int | None
    ):
        archives_to_update: list[Archive] = self.get_archive_entries(
            retrieval_task_id=old_retrieval_job_id
        )
        for archive in archives_to_update:
            archive.retrieval_task_id = new_retrieval_job_id
        self._store.session.commit()

    def get_spring_files_retrieved_before(self, date: datetime):
        return self._store.get_files_retrieved_before(date, tag_names=[SequencingFileTag.SPRING])

    def reset_retrieved_archive_data(self, files_to_reset: list[File]):
        """Resets 'retrieval_task_id' and 'retrieved_at' for all files' corresponding archive entries"""
        for file in files_to_reset:
            LOG.debug(f"Resetting retrieval data for file {file.path}")
            file.archive.retrieval_task_id = None
            file.archive.retrieved_at = None
        self.commit()

    @staticmethod
    def get_files_containing_tags(files: list[File], tags: list[set[str]]) -> list[File]:
        """Return files containing specified tags."""
        filtered_files: list[File] = []
        for file in files:
            file_tags: set[str] = {tag.name for tag in file.tags}
            if any(tag.issubset(file_tags) for tag in tags):
                filtered_files.append(file)
        return filtered_files

    @staticmethod
    def get_files_without_excluded_tags(files: list[File], excluded_tags: list[str]) -> list[File]:
        """Return files without specified tags."""
        filtered_files: list[File] = []
        for file in files:
            file_tags: list[str] = [tag.name for tag in file.tags]
            if not any(excluded_tag in file_tags for excluded_tag in excluded_tags):
                filtered_files.append(file)
        return filtered_files

    def get_files_from_latest_version_containing_tags(
        self, bundle_name: str, tags: list[set[str]], excluded_tags: list[str] | None = None
    ) -> list[File]:
        """
        Return files from the latest version of a bundle matching provided tags. Files containing
        any tag sets specified in the excluded_tags list will be excluded from the output.
        """
        files: list[File] = self.get_files_from_latest_version(bundle_name=bundle_name)
        filtered_files: list[File] = self.get_files_containing_tags(files=files, tags=tags)
        if excluded_tags:
            filtered_files: list[File] = self.get_files_without_excluded_tags(
                files=filtered_files, excluded_tags=excluded_tags
            )
        return filtered_files



import logging
from datetime import datetime
from pathlib import Path

from pydantic import TypeAdapter

from cg.apps.lims import LimsAPI
from cg.constants import Priority
from cg.constants.constants import FileFormat
from cg.exc import CgDataError
from cg.io.controller import WriteFile
from cg.services.analysis_starter.configurator.models.microsalt import MicrosaltConfigContent
from cg.services.sequencing_qc_service import SequencingQCService
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class MicrosaltConfigFileCreator:

    def __init__(self, lims_api: LimsAPI, queries_path: str, store: Store):
        self.lims_api = lims_api
        self.queries_path = queries_path
        self.store = store

    def create(self, case_id: str) -> None:
        LOG.info(f"Creating microSALT config file for {case_id}")
        content: list[MicrosaltConfigContent] = self._get_content(case_id)
        content_adapter = TypeAdapter(list[MicrosaltConfigContent])
        config_case_path: Path = self.get_config_path(case_id)
        WriteFile.write_file_from_content(
            content=content_adapter.dump_python(content),
            file_format=FileFormat.JSON,
            file_path=config_case_path,
        )

    def _get_content(self, case_id: str) -> list[MicrosaltConfigContent]:
        case: Case = self.store.get_case_by_internal_id(case_id)
        return [self._get_content_for_sample(sample) for sample in case.samples]

    def _get_content_for_sample(self, sample: Sample) -> MicrosaltConfigContent:
        """Fill a dict with case config information for one sample"""

        sample_id: str = sample.internal_id
        method_library_prep: str | None = self.lims_api.get_prep_method(sample_id)
        method_sequencing: str | None = self.lims_api.get_sequencing_method(sample_id)
        priority = Priority.research.name if sample.priority == 0 else Priority.standard.name
        passes_sequencing_qc: bool = SequencingQCService.sample_pass_sequencing_qc(sample)

        return MicrosaltConfigContent(
            CG_ID_project=self.lims_api.get_sample_project(sample_id),
            Customer_ID_project=sample.original_ticket,
            CG_ID_sample=sample.internal_id,
            Customer_ID_sample=sample.name,
            organism=self._get_organism(sample),
            priority=priority,
            reference=sample.reference_genome,
            Customer_ID=sample.customer.internal_id,
            application_tag=sample.application_version.application.tag,
            date_arrival=str(sample.received_at or datetime.min),
            date_sequencing=str(sample.last_sequenced_at or datetime.min),
            date_libprep=str(sample.prepared_at or datetime.min),
            method_libprep=method_library_prep or "Not in LIMS",
            method_sequencing=method_sequencing or "Not in LIMS",
            sequencing_qc_passed=passes_sequencing_qc,
        )

    @staticmethod
    def _get_organism(sample: Sample) -> str:
        """
        Gets organism special handling for certain species. Fallback on reference genome.
        """

        if not sample.organism:
            raise CgDataError("Organism missing on Sample")

        organism: str = sample.organism.internal_id.strip()

        if "gonorrhoeae" in organism:
            organism = "Neisseria spp."
        elif "Cutibacterium acnes" in organism:
            organism = "Propionibacterium acnes"

        if organism == "VRE":
            reference_genome = sample.organism.reference_genome
            if reference_genome == "NC_017960.1":
                organism = "Enterococcus faecium"
            elif reference_genome == "NC_004668.1":
                organism = "Enterococcus faecalis"

        return organism

    def get_config_path(self, case_id: str) -> Path:
        return Path(self.queries_path, case_id).with_suffix(".json")



from abc import ABC

from cg.services.analysis_starter.configurator.abstract_model import CaseConfig


class Submitter(ABC):
    def submit(self, case_config: CaseConfig) -> CaseConfig:
        """Submits the case to be run either as a SubProcess or using the Seqera Platform.
        Returns the workflowId if run via the Seqera Platform."""
        pass



import logging
import subprocess

from cg.services.analysis_starter.configurator.models.microsalt import MicrosaltCaseConfig
from cg.services.analysis_starter.configurator.models.mip_dna import MIPDNACaseConfig
from cg.services.analysis_starter.submitters.submitter import Submitter
from cg.services.analysis_starter.submitters.subprocess.commands import WORKFLOW_VERSION_COMMAND_MAP

LOG = logging.getLogger(__name__)

SubprocessCaseConfig = MicrosaltCaseConfig | MIPDNACaseConfig


class SubprocessSubmitter(Submitter):
    def submit(self, case_config: SubprocessCaseConfig) -> SubprocessCaseConfig:
        command: str = case_config.get_start_command()
        LOG.info(f"Running: {command}")
        subprocess.run(
            args=command,
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return case_config

    @staticmethod
    def get_workflow_version(case_config: SubprocessCaseConfig) -> str:
        """
        Calls the workflow to get the workflow version number.
        If fails, returns a placeholder value instead.

        Currently supported workflow: microSALT
        """
        try:
            command: str = WORKFLOW_VERSION_COMMAND_MAP[case_config.workflow]
            command = command.format(**case_config.model_dump())
            LOG.debug(f"Running: {command}")
            result: subprocess.CompletedProcess = subprocess.run(
                args=command,
                shell=True,
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout: str = result.stdout.decode("utf-8").rstrip()
            return stdout.split()[-1]
        except Exception:
            LOG.warning(f"Could not retrieve {case_config.workflow} workflow version!")
            return "0.0.0"



import logging

import rich_click as click

from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.microsalt.microsalt import MicrosaltAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Case

LOG = logging.getLogger(__name__)


class MicrosaltUploadAPI(UploadAPI):
    def __init__(self, config: CGConfig):
        self.analysis_api = MicrosaltAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case: Case, restart: bool) -> None:
        """Uploads MicroSALT analysis data and files."""
        analysis: Analysis = self.status_db.get_latest_completed_analysis_for_case(case.internal_id)
        self.update_upload_started_at(analysis)

        self.upload_files_to_customer_inbox(case)
        self.update_uploaded_at(analysis=analysis)



"""Common CLI workflow functions"""

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.balsamic.base import balsamic
from cg.cli.workflow.balsamic.pon import balsamic_pon
from cg.cli.workflow.balsamic.umi import balsamic_umi
from cg.cli.workflow.fluffy.base import fluffy
from cg.cli.workflow.jasen.base import jasen
from cg.cli.workflow.microsalt.base import microsalt
from cg.cli.workflow.mip_dna.base import mip_dna
from cg.cli.workflow.mip_rna.base import mip_rna
from cg.cli.workflow.mutant.base import mutant
from cg.cli.workflow.nallo.base import nallo
from cg.cli.workflow.raredisease.base import raredisease
from cg.cli.workflow.raw_data.base import raw_data
from cg.cli.workflow.rnafusion.base import rnafusion
from cg.cli.workflow.taxprofiler.base import taxprofiler
from cg.cli.workflow.tomte.base import tomte


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
def workflow():
    """Workflows commands"""


workflow.add_command(balsamic)
workflow.add_command(balsamic_pon)
workflow.add_command(balsamic_umi)
workflow.add_command(fluffy)
workflow.add_command(jasen)
workflow.add_command(microsalt)
workflow.add_command(mip_dna)
workflow.add_command(mip_rna)
workflow.add_command(mutant)
workflow.add_command(nallo)
workflow.add_command(raredisease)
workflow.add_command(raw_data)
workflow.add_command(rnafusion)
workflow.add_command(taxprofiler)
workflow.add_command(tomte)



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.pacbio_long_read.constants import PacbioDeliveryType
from cg.services.orders.validation.order_types.pacbio_long_read.models.sample import PacbioSample


class PacbioOrder(OrderWithSamples):
    delivery_type: PacbioDeliveryType
    samples: list[PacbioSample]

    @property
    def enumerated_samples(self) -> enumerate[PacbioSample]:
        return enumerate(self.samples)



import logging
from datetime import datetime
from enum import Enum

import genologics.entities

from cg.apps.lims import LimsAPI
from cg.store.models import Pool, Sample
from cg.store.store import Store
from cg.utils.date import get_date_days_ago

LOG = logging.getLogger(__name__)


class SampleState(Enum):
    RECEIVED = "received"
    PREPARED = "prepared"
    DELIVERED = "delivered"


class PoolState(Enum):
    RECEIVED = "received"
    DELIVERED = "delivered"


class IncludeOptions(Enum):
    UNSET = "unset"
    NOTINVOICED = "not-invoiced"
    ALL = "all"


class TransferLims(object):
    def __init__(self, status: Store, lims: LimsAPI):
        self.status = status
        self.lims = lims

        self._sample_functions = {
            SampleState.RECEIVED: self.status.get_samples_to_receive,
            SampleState.PREPARED: self.status.get_samples_to_prepare,
            SampleState.DELIVERED: self.status.get_samples_to_deliver,
        }

        self._pool_functions = {
            PoolState.RECEIVED: self.status.get_pools_to_receive,
            PoolState.DELIVERED: self.status.get_all_pools_to_deliver,
        }

        self._date_functions = {
            SampleState.RECEIVED: self.lims.get_received_date,
            SampleState.PREPARED: self.lims.get_prepared_date,
            SampleState.DELIVERED: self.lims.get_delivery_date,
            PoolState.RECEIVED: self.lims.get_received_date,
            PoolState.DELIVERED: self.lims.get_delivery_date,
        }

    def transfer_samples(
        self,
        status_type: SampleState,
        include: str = "unset",
        max_order_age: int | None = None,
        sample_id: str = None,
    ):
        """Transfer information about samples."""

        if sample_id:
            samples: list[Sample] = self.status.get_samples_by_internal_id(internal_id=sample_id)
        else:
            samples: list[Sample] | None = self._get_samples_to_include(
                include=include, status_type=status_type
            )

        if max_order_age:
            order_date_cutoff: datetime = get_date_days_ago(max_order_age * 365)
            LOG.info(f"Remove samples ordered before {str(order_date_cutoff)}")
            samples: list[Sample] = self._filter_out_older_orders(
                samples=samples, order_date_cutoff=order_date_cutoff
            )

        if samples is None:
            LOG.info(f"No samples to process found with {include} {status_type.value}")
            return
        else:
            LOG.info(f"{len(samples)} samples to process")

        for sample_obj in samples:
            lims_date = self._date_functions[status_type](sample_obj.internal_id)
            statusdb_date = getattr(sample_obj, f"{status_type.value}_at")
            if lims_date:
                if statusdb_date and statusdb_date.date() == lims_date:
                    continue

                LOG.info(
                    f"Found new {status_type.value} date for {sample_obj.internal_id}: "
                    f"{lims_date}, old value: {statusdb_date} "
                )

                setattr(sample_obj, f"{status_type.value}_at", lims_date)
                self.status.commit_to_store()
            else:
                LOG.debug(f"no {status_type.value} date found for {sample_obj.internal_id}")

    def _filter_out_older_orders(
        self, samples: list[Sample], order_date_cutoff: datetime
    ) -> list[Sample]:
        sample_within_time_window: list[Sample] = []
        for sample in samples:
            if not self._is_sample_too_old(sample=sample, order_date_cutoff=order_date_cutoff):
                sample_within_time_window.append(sample)
        LOG.debug(f"Filter out {len(samples)-len(sample_within_time_window)} samples")
        return sample_within_time_window

    def _get_samples_to_include(self, include, status_type):
        samples = None
        if include == IncludeOptions.UNSET.value:
            samples = self._get_samples_in_step(status_type)
        elif include == IncludeOptions.NOTINVOICED.value:
            samples = self.status.get_samples_not_invoiced()
        elif include == IncludeOptions.ALL.value:
            samples = self.status.get_samples_not_down_sampled()
        return samples

    def transfer_pools(self, status_type: PoolState):
        """Transfer information about pools."""
        pools = self._pool_functions[status_type]()

        for pool_obj in pools:
            ticket: str = pool_obj.ticket
            number_of_samples: int = self.lims.get_sample_number(projectname=ticket)
            if not self._is_pool_valid(pool_obj, ticket, number_of_samples):
                continue

            samples_in_pool: dict[str, str] | list[genologics.Sample] = self.lims.get_samples(
                projectname=ticket
            )
            for sample_obj in samples_in_pool:
                if not self._is_sample_valid(pool_obj, sample_obj):
                    continue
                status_date = self._date_functions[status_type](sample_obj.id)
                if status_date is None:
                    continue

                LOG.info(f"Found {status_type.value} date for pool id {pool_obj.id}: {status_date}")
                setattr(pool_obj, f"{status_type.value}_at", status_date)
                self.status.session.commit()
                break

    def _get_samples_in_step(self, status_type) -> list[Sample]:
        return self._sample_functions[status_type]()

    @staticmethod
    def _is_pool_valid(pool_obj: Pool, ticket: str, number_of_samples: int) -> bool:
        """Checks if a pool object can be transferred. A pool needs to have a ticket number and at least one sample"""

        if ticket is None:
            LOG.warning(f"No ticket number found for pool with order number {pool_obj.order}.")
            return False
        if number_of_samples == 0:
            LOG.warning(f"No samples found for pool with ticket number {ticket}.")
            return False
        return True

    @staticmethod
    def _is_sample_valid(pool_obj: Pool, sample_obj: genologics.entities.Sample) -> bool:
        """Checks if a sample can have the status date set. A sample needs to have a udf "pool name" that matches the
        name of the pool object it's part of"""
        if sample_obj.udf.get("pool name") is None:
            LOG.warning(
                f"No pool name found for sample {sample_obj} (sample name {sample_obj.name}, project {sample_obj.project.name})"
            )
            return False

        sample_pool_name = sample_obj.udf["pool name"]
        if sample_pool_name != pool_obj.name:
            LOG.warning(
                f"The udf 'pool name' of the sample ({sample_pool_name}) does not match the name of the pool ({pool_obj.name}))"
            )
            return False
        return True

    @staticmethod
    def _is_sample_too_old(sample: Sample, order_date_cutoff: datetime) -> bool:
        return sample.ordered_at < order_date_cutoff



from cg.models.qc_metrics import QCMetrics


class TaxprofilerQCMetrics(QCMetrics):
    """Taxprofiler QC metrics."""

    after_filtering_gc_content: float
    after_filtering_read1_mean_length: float
    after_filtering_read2_mean_length: float
    after_filtering_total_reads: float
    average_length: float
    pct_duplication: float
    raw_total_sequences: float
    reads_mapped: float



from typing import Annotated

from pydantic import BeforeValidator, Field

from cg.constants import SexOptions
from cg.models.qc_metrics import QCMetrics


def convert_sex(plink_sex: float) -> SexOptions:
    if plink_sex == 2:
        return SexOptions.FEMALE
    elif plink_sex == 1:
        return SexOptions.MALE
    elif plink_sex == 0:
        return SexOptions.UNKNOWN
    else:
        raise NotImplementedError


class NalloQCMetrics(QCMetrics):
    """Nallo QC metrics."""

    avg_sequence_length: float | None
    coverage_bases: float | None
    median_coverage: float | None
    percent_duplicates: float | None
    predicted_sex: Annotated[SexOptions, BeforeValidator(convert_sex)] = Field(alias="somalier_sex")



from pathlib import Path

from cg.constants.scout import ScoutExportFileName
from cg.services.analysis_starter.configurator.extensions.pipeline_extension import (
    PipelineExtension,
)
from cg.services.analysis_starter.configurator.file_creators.gene_panel import GenePanelFileCreator


class NalloExtension(PipelineExtension):
    def __init__(self, gene_panel_file_creator: GenePanelFileCreator):
        self.gene_panel_file_creator = gene_panel_file_creator

    def configure(self, case_id: str, case_run_directory: Path) -> None:
        self.gene_panel_file_creator.create(
            case_id=case_id,
            file_path=self._get_gene_panel_file_path(case_run_directory),
            double_hashtag_filtering=True,
        )

    def do_required_files_exist(self, case_run_directory: Path) -> bool:
        gene_panel_file_path: Path = self._get_gene_panel_file_path(case_run_directory)
        return gene_panel_file_path.is_file()

    @staticmethod
    def _get_gene_panel_file_path(case_run_directory: Path) -> Path:
        return case_run_directory.joinpath(ScoutExportFileName.PANELS_TSV)



from cg.services.orders.validation.rules.sample.rules import (
    validate_application_compatibility,
    validate_application_exists,
    validate_applications_not_archived,
    validate_index_number_in_range,
    validate_index_number_required,
    validate_pools_contain_one_application,
    validate_pools_contain_one_priority,
    validate_sample_names_available,
    validate_sample_names_unique,
    validate_volume_interval,
    validate_volume_required,
    validate_well_position_rml_format,
    validate_well_positions_required_rml,
)

RML_SAMPLE_RULES: list[callable] = [
    validate_application_compatibility,
    validate_application_exists,
    validate_applications_not_archived,
    validate_index_number_required,
    validate_index_number_in_range,
    validate_pools_contain_one_application,
    validate_pools_contain_one_priority,
    validate_sample_names_available,
    validate_sample_names_unique,
    validate_volume_interval,
    validate_volume_required,
    validate_well_position_rml_format,
    validate_well_positions_required_rml,
]



from cg.models.orders.sample_base import PriorityEnum, SexEnum
from cg.services.orders.validation.models.sample import Sample


class PacbioSample(Sample):
    concentration_ng_ul: float | None = None
    priority: PriorityEnum
    quantity: int | None = None
    require_qc_ok: bool = False
    sex: SexEnum
    source: str
    tumour: bool = False



from datetime import datetime
from pathlib import Path
from typing import Any, cast

from cg.constants.priority import SlurmQos
from cg.exc import SeqeraError
from cg.io.yaml import read_yaml, write_yaml_stream
from cg.services.analysis_starter.configurator.models.nextflow import NextflowCaseConfig
from cg.services.analysis_starter.submitters.seqera_platform.dtos import (
    LaunchRequest,
    WorkflowLaunchRequest,
)
from cg.services.analysis_starter.submitters.seqera_platform.seqera_platform_client import (
    SeqeraPlatformClient,
)
from cg.services.analysis_starter.submitters.submitter import Submitter


class SeqeraPlatformSubmitter(Submitter):

    def __init__(self, client: SeqeraPlatformClient, compute_environment_ids: dict[SlurmQos, str]):
        self.client: SeqeraPlatformClient = client
        self.compute_environment_ids: dict[str, str] = compute_environment_ids

    def submit(self, case_config: NextflowCaseConfig) -> NextflowCaseConfig:
        """Starts a case and returns the workflow id for the job."""
        new_case_config: NextflowCaseConfig = case_config.model_copy()
        run_request: WorkflowLaunchRequest = self._create_launch_request(new_case_config)

        response: dict = self.client.launch_workflow(run_request)
        if not (workflow_id := response.get("workflowId")):
            raise SeqeraError(f"workflowId missing from response: {response}")
        new_case_config.workflow_id = workflow_id

        workflow_response: dict = self.client.get_workflow(cast(str, new_case_config.workflow_id))
        if not (session_id := self._get_session_id(workflow_response)):
            raise SeqeraError(f"No sessionId found in response: {workflow_response}")
        new_case_config.session_id = session_id

        return new_case_config

    @staticmethod
    def _get_session_id(workflow_response: dict[str, Any]) -> str | None:
        workflow: dict | None = workflow_response.get("workflow")
        if workflow and (session_id := workflow.get("sessionId")):
            return session_id
        else:
            return None

    def _create_launch_request(self, case_config: NextflowCaseConfig) -> WorkflowLaunchRequest:
        parameters: dict = read_yaml(Path(case_config.params_file))
        parameter_stream: str = write_yaml_stream(parameters)
        launch_request = LaunchRequest(
            computeEnvId=self.compute_environment_ids[case_config.case_priority],
            configProfiles=case_config.config_profiles,
            configText=f"includeConfig '{case_config.nextflow_config_file}'",
            paramsText=parameter_stream,
            pipeline=case_config.pipeline_repository,
            preRunScript=case_config.pre_run_script,
            resume=case_config.resume,
            revision=case_config.revision,
            runName=self._create_run_name(case_id=case_config.case_id, resume=case_config.resume),
            sessionId=case_config.session_id,
            workDir=case_config.work_dir,
        )
        return WorkflowLaunchRequest(launch=launch_request)

    @staticmethod
    def _create_run_name(case_id: str, resume: bool) -> str:
        if resume:
            return f"{case_id}_resumed_{datetime.now().strftime('%Y-%m-%d_%H-%M')}"
        else:
            return case_id



"""CLI support to create config and/or start BALSAMIC."""

import logging
from typing import cast

import rich_click as click
from pydantic.v1 import ValidationError

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.balsamic.options import OPTION_PANEL_BED, OPTION_WORKFLOW_PROFILE
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, resolve_compression
from cg.cli.workflow.utils import validate_force_store_option
from cg.constants import Workflow
from cg.constants.cli_options import COMMENT, DRY_RUN, FORCE
from cg.exc import CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.analysis_starter import AnalysisStarter
from cg.services.analysis_starter.configurator.implementations.balsamic import BalsamicConfigurator
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.factories.starter_factory import AnalysisStarterFactory
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def balsamic(context: click.Context):
    """Cancer analysis workflow"""
    AnalysisAPI.get_help(context)

    config = context.obj
    context.obj.meta_apis["analysis_api"] = BalsamicAnalysisAPI(config=config)


balsamic.add_command(resolve_compression)


@balsamic.command("report-deliver")
@ARGUMENT_CASE_ID
@DRY_RUN
@FORCE
@click.pass_obj
def report_deliver(context: CGConfig, case_id: str, dry_run: bool, force: bool):
    """Create a Housekeeper deliverables file for a given case ID."""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    try:
        analysis_api.status_db.verify_case_exists(case_id)
        analysis_api.verify_case_config_file_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.trailblazer_api.verify_latest_analysis_is_completed(
            case_id=case_id, force=force
        )
        analysis_api.report_deliver(case_id=case_id, dry_run=dry_run)
    except CgError as error:
        LOG.error(f"Could not create report file: {error}")
        raise click.Abort()
    except Exception as error:
        LOG.error(f"Could not create report file: {error}")
        raise click.Abort()


@balsamic.command("store-housekeeper")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_obj
def store_housekeeper(
    context: CGConfig, case_id: str, comment: str | None, dry_run: bool, force: bool
):
    """Store a finished analysis in Housekeeper and StatusDB."""

    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db

    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.verify_case_config_file_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.verify_deliverables_file_exists(case_id=case_id)
        _, version = analysis_api.create_housekeeper_bundle(
            case_id=case_id, dry_run=dry_run, force=force
        )
        analysis_api.update_analysis_as_completed_statusdb(
            case_id=case_id, hk_version_id=version.id, comment=comment, dry_run=dry_run, force=force
        )
        analysis_api.set_statusdb_action(case_id=case_id, action=None, dry_run=dry_run)
    except ValidationError as error:
        LOG.warning("Deliverables file is malformed")
        raise error
    except CgError as error:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {error}")
        raise click.Abort()
    except Exception as error:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {error}!")
        housekeeper_api.rollback()
        status_db.session.rollback()
        raise click.Abort()


@balsamic.command("store")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_context
def store(context: click.Context, case_id: str, comment: str | None, dry_run: bool, force: bool):
    """Generate Housekeeper report for CASE ID and store in Housekeeper"""
    LOG.info(f"Storing analysis for {case_id}")
    validate_force_store_option(force=force, comment=comment)
    context.invoke(report_deliver, case_id=case_id, dry_run=dry_run, force=force)
    context.invoke(store_housekeeper, case_id=case_id, comment=comment, force=force)


@balsamic.command("store-available")
@DRY_RUN
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """Store bundles for all finished analyses in Housekeeper"""

    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    was_successful: bool = True
    for case_obj in analysis_api.get_cases_to_store():
        LOG.info(f"Storing deliverables for {case_obj.internal_id}")
        try:
            context.invoke(store, case_id=case_obj.internal_id, dry_run=dry_run)
        except Exception as exception_object:
            LOG.error(f"Error storing {case_obj.internal_id}: {exception_object}")
            was_successful = False
    if not was_successful:
        raise click.Abort()


@balsamic.command("config-case")
@OPTION_PANEL_BED
@ARGUMENT_CASE_ID
@click.pass_obj
def config_case(cg_config: CGConfig, case_id: str, panel_bed: str | None):
    """Configure a Balsamic case so that it is ready to be run.

    \b
    Creates the case config file:
        - CASE_ID.json
    """
    factory = ConfiguratorFactory(cg_config)
    configurator = cast(BalsamicConfigurator, factory.get_configurator(Workflow.BALSAMIC))
    configurator.configure(case_id=case_id, panel_bed=panel_bed)


@balsamic.command("run")
@OPTION_WORKFLOW_PROFILE
@ARGUMENT_CASE_ID
@click.pass_obj
def run(cg_config: CGConfig, case_id: str, workflow_profile: click.Path | None):
    """
    Run a preconfigured Balsamic case.

    \b
    Assumes that case config file exist in the case run directory:
        - CASE_ID.json
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(Workflow.BALSAMIC)
    analysis_starter.run(case_id=case_id, workflow_profile=workflow_profile)


@balsamic.command("start")
@OPTION_PANEL_BED
@OPTION_WORKFLOW_PROFILE
@ARGUMENT_CASE_ID
@click.pass_obj
def start(
    cg_config: CGConfig,
    case_id: str,
    panel_bed: str | None,
    workflow_profile: click.Path | None,
):
    """
    Starts a Balsamic cases.

    \b
    Configures the case and creates the case config file:
        - CASE_ID.json
    and submits the job to slurm.
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(Workflow.BALSAMIC)
    analysis_starter.start(case_id=case_id, workflow_profile=workflow_profile, panel_bed=panel_bed)


@balsamic.command("start-available")
@click.pass_obj
def start_available(cg_config: CGConfig):
    """
    Starts all available Balsamic cases.

    \b
    Configures the individual case and creates its case config file:
        - CASE_ID.json
    and submits the job to slurm.
    """
    LOG.info("Starting Balsamic workflow for all available cases.")
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter = factory.get_analysis_starter_for_workflow(Workflow.BALSAMIC)
    succeeded: bool = analysis_starter.start_available()
    if not succeeded:
        raise click.Abort



from cg.services.delivery_message.messages.analysis_scout_message import AnalysisScoutMessage
from cg.services.delivery_message.messages.covid_message import CovidMessage
from cg.services.delivery_message.messages.fastq_analysis_scout_message import (
    FastqAnalysisScoutMessage,
)
from cg.services.delivery_message.messages.fastq_message import FastqMessage
from cg.services.delivery_message.messages.fastq_scout_message import FastqScoutMessage
from cg.services.delivery_message.messages.microsalt_message import MicrosaltMessage
from cg.services.delivery_message.messages.rna_delivery_message import (
    RNAAnalysisStrategy,
    RNADeliveryMessage,
    RNAFastqAnalysisStrategy,
    RNAFastqStrategy,
    RNAScoutStrategy,
    RNAUploadMessageStrategy,
)
from cg.services.delivery_message.messages.scout_message import ScoutMessage
from cg.services.delivery_message.messages.statina_message import StatinaMessage

__all__ = [
    "AnalysisScoutMessage",
    "CovidMessage",
    "FastqAnalysisScoutMessage",
    "FastqMessage",
    "FastqScoutMessage",
    "MicrosaltMessage",
    "RNAAnalysisStrategy",
    "RNADeliveryMessage",
    "RNAFastqAnalysisStrategy",
    "RNAFastqStrategy",
    "RNAScoutStrategy",
    "RNAUploadMessageStrategy",
    "ScoutMessage",
    "StatinaMessage",
]



"""Module for Tomte Analysis API."""

import logging
from pathlib import Path

from cg.constants import Workflow
from cg.constants.constants import GenomeVersion
from cg.constants.nf_analysis import TOMTE_METRIC_CONDITIONS
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase
from cg.models.tomte.tomte import TomteQCMetrics
from cg.resources import TOMTE_BUNDLE_FILENAMES_PATH

LOG = logging.getLogger(__name__)


class TomteAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Tomte processes and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.TOMTE,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.tomte.root
        self.workflow_bin_path: str = config.tomte.workflow_bin_path
        self.profile: str = config.tomte.profile
        self.conda_env: str = config.tomte.conda_env
        self.conda_binary: str = config.tomte.conda_binary
        self.platform: str = config.tomte.platform
        self.params: str = config.tomte.params
        self.workflow_config_path: str = config.tomte.config
        self.resources: str = config.tomte.resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.tomte.tower_workflow
        self.account: str = config.tomte.slurm.account
        self.email: str = config.tomte.slurm.mail_user
        self.compute_env_base: str = config.tomte.compute_env
        self.revision: str = config.tomte.revision
        self.nextflow_binary_path: str = config.tomte.binary_path

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return path to bundle template."""
        return TOMTE_BUNDLE_FILENAMES_PATH

    def get_genome_build(self, case_id: str) -> str:
        return GenomeVersion.HG38

    def get_workflow_metrics(self, metric_id: str) -> dict:
        return TOMTE_METRIC_CONDITIONS

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        qc_metrics_model = TomteQCMetrics
        return super().parse_analysis(
            qc_metrics_raw=qc_metrics_raw, qc_metrics_model=qc_metrics_model, **kwargs
        )



from collections.abc import Iterator

from cg.constants.constants import Strandedness
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.creator import (
    NextflowFastqSampleSheetCreator,
)
from cg.store.models import Case, Sample

HEADERS: list[str] = ["sample", "fastq_1", "fastq_2", "strandedness"]


class RNAFusionSampleSheetCreator(NextflowFastqSampleSheetCreator):

    def _get_content(self, case_id: str) -> list[list[str]]:
        content: list[list[str]] = [HEADERS]
        case: Case = self.store.get_case_by_internal_id(case_id)
        for sample in case.samples:
            content.extend(self._get_sample_sheet_content_per_sample(sample))
        return content

    def _get_sample_sheet_content_per_sample(self, sample: Sample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        paired_fastq_files: Iterator[tuple[str, str]] = self._get_paired_read_paths(sample)
        content: list[list[str]] = []
        for fastq_forward_read_path, fastq_reverse_read_path in paired_fastq_files:
            content.append(
                [
                    sample.internal_id,
                    fastq_forward_read_path,
                    fastq_reverse_read_path,
                    Strandedness.REVERSE,
                ]
            )
        return content



import logging
from pathlib import Path

from cg.apps.lims import LimsAPI
from cg.constants import FileExtensions
from cg.constants.scout import ScoutExportFileName
from cg.exc import CgDataError
from cg.io.csv import write_csv
from cg.io.yaml import read_yaml, write_yaml_nextflow_style
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.models import (
    RarediseaseParameters,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.utils import (
    replace_values_in_params_file,
)
from cg.store.models import BedVersion, Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class RarediseaseParamsFileCreator(ParamsFileCreator):

    def __init__(self, store: Store, lims: LimsAPI, params: str):
        super().__init__(params)
        self.store = store
        self.lims = lims

    def create(self, case_id: str, file_path: Path, sample_sheet_path: Path) -> None:
        LOG.debug(f"Creating params file for case {case_id}")
        content: dict = self._get_content(
            case_id=case_id, case_path=file_path.parent, sample_sheet_path=sample_sheet_path
        )
        write_yaml_nextflow_style(file_path=file_path, content=content)

    def _get_content(self, case_id: str, case_path: Path, sample_sheet_path: Path) -> dict:
        """Return the merged dictionary with case-specific parameters and workflow parameters."""
        case_parameters: dict = self._get_case_parameters(
            case_id=case_id, case_path=case_path, sample_sheet_path=sample_sheet_path
        ).model_dump()
        workflow_parameters: dict = read_yaml(self.params)
        if duplicate_keys := set(case_parameters.keys()) & set(workflow_parameters.keys()):
            raise CgDataError(f"Duplicate parameter keys found: {duplicate_keys}")
        parameters: dict = case_parameters | workflow_parameters
        curated_parameters: dict = replace_values_in_params_file(parameters)
        return curated_parameters

    def _get_case_parameters(
        self, case_id: str, case_path: Path, sample_sheet_path: Path
    ) -> RarediseaseParameters:
        """Return case-specific parameters for the analysis."""
        analysis_type: str = self._get_arbitrary_prep_category_in_case(case_id)
        target_bed_file: str = self._get_target_bed_from_lims(case_id)
        sample_mapping_file: Path = self._create_sample_mapping_file(
            case_id=case_id, case_path=case_path
        )
        return RarediseaseParameters(
            input=sample_sheet_path,
            outdir=case_path,
            analysis_type=analysis_type,
            target_bed_file=target_bed_file,
            save_mapped_as_cram=True,
            vcfanno_extra_resources=f"{case_path}/{ScoutExportFileName.MANAGED_VARIANTS}",
            vep_filters_scout_fmt=f"{case_path}/{ScoutExportFileName.PANELS}",
            sample_id_map=sample_mapping_file,
        )

    def _get_arbitrary_prep_category_in_case(self, case_id: str) -> str:
        """
        Returns prep category. Assumes all case samples have the same
        prep category.
        """
        sample: Sample = self.store.get_samples_by_case_id(case_id=case_id)[0]
        return sample.prep_category

    def _get_target_bed_from_lims(self, case_id: str) -> str:
        """
        Get target bed filename from LIMS. Return an empty string if no target bed is found in LIMS.
        Raises:
            CgDataError: if the bed target capture version is not found in StatusDB.
        """
        case: Case = self.store.get_case_by_internal_id_strict(internal_id=case_id)
        sample: Sample = case.samples[0]
        target_bed_shortname: str | None = self.lims.capture_kit(
            sample.from_sample or sample.internal_id
        )
        if target_bed_shortname:
            bed_version: BedVersion = self.store.get_bed_version_by_short_name_strict(
                target_bed_shortname
            )
            return bed_version.filename
        else:
            return ""

    def _create_sample_mapping_file(self, case_id: str, case_path: Path) -> Path:
        """Create a sample mapping file for the case and returns its path."""
        file_path = Path(case_path, f"{case_id}_customer_internal_mapping").with_suffix(
            FileExtensions.CSV
        )
        content: list[list[str]] = [["customer_id", "internal_id"]]
        case: Case = self.store.get_case_by_internal_id_strict(internal_id=case_id)
        content.extend([sample.name, sample.internal_id] for sample in case.samples)
        write_csv(file_path=file_path, content=content)
        return file_path



"""Delivery report specific command options."""

import rich_click as click

from cg.constants import REPORT_SUPPORTED_WORKFLOW

ARGUMENT_CASE_ID = click.argument(
    "case_id",
    required=False,
    type=str,
)

OPTION_WORKFLOW = click.option(
    "--workflow",
    type=click.Choice(REPORT_SUPPORTED_WORKFLOW),
    help="Limit delivery report generation to a specific workflow",
)



from pathlib import Path

from cg.constants.constants import WorkflowManager
from cg.io.yaml import read_yaml
from cg.services.analysis_starter.configurator.models.mip_dna import MIPDNACaseConfig
from cg.services.analysis_starter.tracker.tracker import Tracker


class MIPDNATracker(Tracker):

    def _workflow_manager(self) -> WorkflowManager:
        return WorkflowManager.Slurm

    def _get_job_ids_path(self, case_id: str) -> Path:
        return Path(self._get_out_dir_path(case_id), "slurm_job_ids.yaml")

    def _get_out_dir_path(self, case_id: str) -> Path:
        return Path(self.workflow_root, case_id, "analysis")

    def _get_sample_info_path(self, case_id: str) -> Path:
        return Path(self.workflow_root, case_id, "analysis", f"{case_id}_qc_sample_info.yaml")

    def _get_workflow_version(self, case_config: MIPDNACaseConfig) -> str:
        sample_info_raw: dict = read_yaml(file_path=self._get_sample_info_path(case_config.case_id))
        return sample_info_raw["mip_version"]



from cg.server.endpoints.sequencing_run.dtos import (
    PacbioSequencingRunDTO,
    PacbioSequencingRunResponse,
    PacbioSequencingRunUpdateRequest,
    PacbioSmrtCellMetricsDTO,
    PacbioSmrtCellMetricsResponse,
)
from cg.store.models import PacbioSMRTCellMetrics
from cg.store.store import Store


class PacbioSequencingRunsService:
    def __init__(self, store: Store):
        self.store = store

    def get_sequencing_runs_by_name(self, run_name: str) -> PacbioSmrtCellMetricsResponse:
        runs: list[PacbioSmrtCellMetricsDTO] = []
        db_runs: list[PacbioSMRTCellMetrics] = self.store.get_pacbio_smrt_cell_metrics_by_run_name(
            run_name
        )
        for db_run in db_runs:
            run_dict = db_run.to_dict()
            run_dict["internal_id"] = db_run.device.internal_id
            run_dict["run_name"] = db_run.sequencing_run.run_name
            run = PacbioSmrtCellMetricsDTO.model_validate(run_dict)
            runs.append(run)
        return PacbioSmrtCellMetricsResponse(runs=runs)

    def get_sequencing_runs(self, page: int = 0, page_size: int = 0) -> PacbioSequencingRunResponse:
        db_runs, total_count = self.store.get_pacbio_sequencing_runs(page=page, page_size=page_size)
        runs: list[PacbioSequencingRunDTO] = []
        for db_run in db_runs:
            run = PacbioSequencingRunDTO(
                id=db_run.id,
                run_name=db_run.run_name,
                comment=db_run.comment,
                processed=db_run.processed,
            )
            runs.append(run)
        return PacbioSequencingRunResponse(pacbio_sequencing_runs=runs, total_count=total_count)

    def update_sequencing_run(self, update_request: PacbioSequencingRunUpdateRequest) -> None:
        if update_request.comment is not None:
            self.store.update_pacbio_sequencing_run_comment(
                id=update_request.id, comment=update_request.comment
            )

        if update_request.processed is not None:
            self.store.update_pacbio_sequencing_run_processed(
                id=update_request.id, processed=update_request.processed
            )



from pydantic import BaseModel

from cg.services.sample_run_metrics_service.dtos import PacbioSequencingMetricsDTO


class PacbioSequencingMetricsRequest(BaseModel):
    sample_id: str | None = None
    smrt_cell_ids: list[str] | None = None


class PacbioSequencingMetricsResponse(BaseModel):
    metrics: list[PacbioSequencingMetricsDTO] = []



"""Module for building the rsync command to send files to customer inbox on the delivery server."""

import datetime as dt
import glob
import logging
from pathlib import Path
from typing import Iterable

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import Workflow
from cg.constants.constants import FileFormat
from cg.constants.delivery import INBOX_NAME
from cg.constants.priority import SlurmAccount, SlurmQos, TrailblazerPriority
from cg.constants.slurm import (
    SLURM_UPLOAD_EXCLUDED_COMPUTE_NODES,
    SLURM_UPLOAD_MAX_HOURS,
    SLURM_UPLOAD_MEMORY,
    SLURM_UPLOAD_TASKS,
)
from cg.constants.tb import AnalysisType
from cg.exc import CgError
from cg.io.controller import WriteFile
from cg.models.slurm.sbatch import Sbatch
from cg.services.deliver_files.rsync.models import RsyncDeliveryConfig
from cg.services.deliver_files.rsync.sbatch_commands import (
    COVID_REPORT_RSYNC,
    COVID_RSYNC,
    CREATE_INBOX_COMMAND,
    ERROR_CREATE_INBOX_FUNCTION,
    ERROR_RSYNC_FUNCTION,
    RSYNC_COMMAND,
)
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class DeliveryRsyncService:
    def __init__(
        self,
        delivery_path: str,
        rsync_config: RsyncDeliveryConfig,
        status_db: Store,
    ):
        self.status_db = status_db
        self.delivery_path: str = delivery_path
        self.destination_path: str = rsync_config.destination_path
        self.covid_destination_path: str = rsync_config.covid_destination_path
        self.covid_report_path: str = rsync_config.covid_report_path
        self.base_path: Path = Path(rsync_config.base_path)
        self.account: str = rsync_config.account
        self.log_dir: Path = Path(rsync_config.base_path)
        self.mail_user: str = rsync_config.mail_user
        self.workflow: str = Workflow.RSYNC

    @property
    def slurm_quality_of_service(self) -> SlurmQos:
        """Return the slurm quality of service depending on the slurm account."""
        return SlurmQos.HIGH if self.account == SlurmAccount.PRODUCTION else SlurmQos.LOW

    @property
    def trailblazer_priority(self) -> TrailblazerPriority:
        """Return the trailblazer priority depending on the slurm account."""
        return (
            TrailblazerPriority.HIGH
            if self.account == SlurmAccount.PRODUCTION
            else TrailblazerPriority.LOW
        )

    @property
    def trailblazer_config_path(self) -> Path:
        """Return Path to trailblazer config."""
        return self.log_dir / "slurm_job_ids.yaml"

    @property
    def rsync_processes(self) -> Iterable[Path]:
        """Yield existing rsync processes."""
        yield from self.base_path.iterdir()

    @staticmethod
    def format_covid_destination_path(
        covid_destination_path: str, customer_internal_id: str
    ) -> str:
        """Return destination path of covid report."""
        return covid_destination_path % customer_internal_id

    @staticmethod
    def get_trailblazer_config(slurm_job_id: int) -> dict[str, list[str]]:
        """Return dictionary of slurm job IDs."""
        return {"jobs": [str(slurm_job_id)]}

    @staticmethod
    def write_trailblazer_config(content: dict, config_path: Path, dry_run: bool = False) -> None:
        """Write slurm job IDs to a .YAML file used as the trailblazer config."""
        if dry_run:
            LOG.info(f"Dry-run: it would write slurm jobs to {config_path.as_posix()}")
            return
        LOG.info(f"Writing slurm jobs to {config_path.as_posix()}")
        WriteFile.write_file_from_content(
            content=content, file_format=FileFormat.YAML, file_path=config_path
        )

    @staticmethod
    def process_ready_to_clean(before: dt.datetime, process: Path) -> bool:
        """Return True if analysis is old enough to be cleaned."""

        ctime: dt.datetime = dt.datetime.fromtimestamp(process.stat().st_ctime)

        return before > ctime

    def concatenate_rsync_commands(
        self,
        folder_list: set[Path],
        source_and_destination_paths: dict[str, Path],
        ticket: str,
        case: Case,
    ) -> str:
        """Concatenates the rsync commands for each folder to be transferred."""
        command: str = ""
        for folder in folder_list:
            source_path = Path(source_and_destination_paths["delivery_source_path"], folder.name)
            destination_path: Path = Path(
                source_and_destination_paths["rsync_destination_path"], ticket
            )

            command += RSYNC_COMMAND.format(
                source_path=source_path, destination_path=destination_path
            )
        if case.data_analysis == Workflow.MUTANT:
            covid_report_path: str = self.format_covid_report_path(case=case, ticket=ticket)
            covid_destination_path: str = self.format_covid_destination_path(
                self.covid_destination_path, customer_internal_id=case.customer.internal_id
            )
            command += COVID_REPORT_RSYNC.format(
                covid_report_path=covid_report_path,
                covid_destination_path=covid_destination_path,
                log_dir=self.log_dir,
            )
        return command

    def set_log_dir(self, folder_prefix: str) -> None:
        if folder_prefix not in str(self.log_dir.as_posix()):
            timestamp = dt.datetime.now()
            timestamp_str = timestamp.strftime("%y%m%d_%H_%M_%S_%f")
            folder_name = Path("_".join([folder_prefix, timestamp_str]))
            LOG.info(f"Setting log dir to: {self.base_path / folder_name}")
            self.log_dir: Path = self.base_path / folder_name

    def get_all_cases_from_ticket(self, ticket: str) -> list[Case]:
        return self.status_db.get_cases_by_ticket_id(ticket_id=ticket)

    def get_source_and_destination_paths(
        self, ticket: str, customer_internal_id: str
    ) -> dict[str, Path]:
        """Return the source and destination paths."""
        source_and_destination_paths: dict[str, Path] = {
            "delivery_source_path": Path(
                self.delivery_path, customer_internal_id, INBOX_NAME, ticket
            ),
            "rsync_destination_path": Path(self.destination_path, customer_internal_id, INBOX_NAME),
        }
        return source_and_destination_paths

    def add_to_trailblazer_api(
        self, tb_api: TrailblazerAPI, slurm_job_id: int, ticket: str, dry_run: bool
    ) -> None:
        """Add rsync process to trailblazer."""
        if dry_run:
            return
        self.write_trailblazer_config(
            content=self.get_trailblazer_config(slurm_job_id),
            config_path=self.trailblazer_config_path,
        )
        tb_api.add_pending_analysis(
            case_id=ticket,
            analysis_type=AnalysisType.OTHER,
            config_path=self.trailblazer_config_path.as_posix(),
            out_dir=self.log_dir.as_posix(),
            priority=self.trailblazer_priority,
            email=self.mail_user,
            workflow=Workflow.RSYNC,
            ticket=ticket,
        )

    def format_covid_report_path(self, case: Case, ticket: str) -> str:
        """Return a formatted of covid report path."""
        covid_report_options: list[str] = glob.glob(
            self.covid_report_path % (case.internal_id, ticket)
        )
        if not covid_report_options:

            raise FileNotFoundError(
                f"No report file could be found with path"
                f" {self.covid_report_path % (case.internal_id, ticket)}!"
            )
        return covid_report_options[0]

    def create_log_dir(self, dry_run: bool) -> None:
        """Create log dir."""
        log_dir: Path = self.log_dir
        LOG.info(f"Creating folder: {log_dir}")
        if log_dir.exists():
            LOG.warning(f"Could not create {log_dir}, this folder already exist")
        elif dry_run:
            LOG.info(f"Would have created path {log_dir}, but this is a dry run")
        else:
            log_dir.mkdir(parents=True, exist_ok=True)

    def run_rsync_for_case(self, case: Case, dry_run: bool, folders_to_deliver: set[Path]) -> int:
        """Submit Rsync commands for delivering a case to the delivery server and return the job id."""
        ticket: str | None = case.latest_ticket
        if not ticket:
            raise CgError(f"Could not find ticket for case {case.internal_id}")

        source_and_destination_paths: dict[str, Path] = self.get_source_and_destination_paths(
            ticket=ticket, customer_internal_id=case.customer.internal_id
        )

        self.set_log_dir(folder_prefix=case.internal_id)
        self.create_log_dir(dry_run=dry_run)

        folder_creation_job_id: int = self._create_remote_ticket_inbox(
            dry_run=dry_run,
            ticket=ticket,
            source_and_destination_paths=source_and_destination_paths,
        )

        return self._deliver_folder_contents(
            case=case,
            dry_run=dry_run,
            folder_creation_job_id=folder_creation_job_id,
            folders_to_deliver=folders_to_deliver,
            ticket=ticket,
            source_and_destination_paths=source_and_destination_paths,
        )

    def run_rsync_for_ticket(self, ticket: str, dry_run: bool) -> int:
        """Runs rsync of a whole ticket folder to the delivery server."""
        self.set_log_dir(folder_prefix=ticket)
        self.create_log_dir(dry_run=dry_run)
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        if not cases:
            LOG.warning(f"Could not find any cases for ticket {ticket}")
            raise CgError()
        customer_internal_id: str = cases[0].customer.internal_id
        source_and_destination_paths: dict[str, Path] = self.get_source_and_destination_paths(
            ticket=ticket, customer_internal_id=customer_internal_id
        )
        if cases[0].data_analysis == Workflow.MUTANT:
            LOG.info("Delivering report for SARS-COV-2 analysis")
            commands = COVID_RSYNC.format(
                source_path=source_and_destination_paths["delivery_source_path"],
                destination_path=source_and_destination_paths["rsync_destination_path"],
                covid_report_path=self.format_covid_report_path(case=cases[0], ticket=ticket),
                covid_destination_path=self.format_covid_destination_path(
                    self.covid_destination_path, customer_internal_id=customer_internal_id
                ),
                log_dir=self.log_dir,
            )
        else:
            commands = RSYNC_COMMAND.format(
                source_path=source_and_destination_paths["delivery_source_path"],
                destination_path=source_and_destination_paths["rsync_destination_path"],
            )

        return self._generate_and_submit_sbatch(
            commands=commands,
            dry_run=dry_run,
            error_command=ERROR_RSYNC_FUNCTION,
            exclude=SLURM_UPLOAD_EXCLUDED_COMPUTE_NODES,
            job_name=f"{ticket}_rsync",
        )

    def _create_remote_ticket_inbox(
        self, dry_run: bool, ticket: str, source_and_destination_paths: dict[str, Path]
    ) -> int:
        host, inbox_path = (
            source_and_destination_paths["rsync_destination_path"].as_posix().split(":")
        )
        inbox_path: str = f"{inbox_path}/{ticket}"
        job_name: str = f"{ticket}_create_inbox"

        commands: str = CREATE_INBOX_COMMAND.format(host=host, inbox_path=inbox_path)
        error_command = ERROR_CREATE_INBOX_FUNCTION

        return self._generate_and_submit_sbatch(
            commands=commands, error_command=error_command, job_name=job_name, dry_run=dry_run
        )

    def _deliver_folder_contents(
        self,
        case: Case,
        folder_creation_job_id: int,
        folders_to_deliver: set[Path],
        ticket: str,
        source_and_destination_paths: dict[str, Path],
        dry_run: bool = False,
    ) -> int:
        """Submit the job that rsyncs case data to the delivery server and return the job id."""
        command: str = self.concatenate_rsync_commands(
            folder_list=folders_to_deliver,
            source_and_destination_paths=source_and_destination_paths,
            ticket=ticket,
            case=case,
        )

        return self._generate_and_submit_sbatch(
            commands=command,
            dry_run=dry_run,
            error_command=ERROR_RSYNC_FUNCTION,
            exclude=SLURM_UPLOAD_EXCLUDED_COMPUTE_NODES,
            job_name=f"{case.internal_id}_rsync",
            dependency=f"--dependency=afterok:{folder_creation_job_id}",
        )

    def _generate_and_submit_sbatch(
        self,
        job_name: str,
        commands: str,
        error_command: str,
        dry_run: bool = False,
        dependency: str | None = None,
        exclude: str | None = None,
    ) -> int:
        sbatch_parameters: Sbatch = Sbatch(
            account=self.account,
            commands=commands,
            dependency=dependency,
            email=self.mail_user,
            error=error_command,
            exclude=exclude,
            hours=SLURM_UPLOAD_MAX_HOURS,
            job_name=job_name,
            log_dir=self.log_dir.as_posix(),
            memory=SLURM_UPLOAD_MEMORY,
            number_tasks=SLURM_UPLOAD_TASKS,
            quality_of_service=self.slurm_quality_of_service,
        )

        slurm_api = SlurmAPI()
        slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_content: str = slurm_api.generate_sbatch_content(sbatch_parameters=sbatch_parameters)
        sbatch_path = Path(self.log_dir, f"{job_name}.sh")
        sbatch_number: int = slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        return sbatch_number



import logging

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Workflow
from cg.services.deliver_files.file_fetcher.abstract import FetchDeliveryFilesService
from cg.services.deliver_files.file_fetcher.error_handling import handle_missing_bundle_errors
from cg.services.deliver_files.file_fetcher.exc import NoDeliveryFilesError
from cg.services.deliver_files.file_fetcher.models import (
    DeliveryFiles,
    DeliveryMetaData,
    SampleFile,
)
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class RawDataDeliveryFileFetcher(FetchDeliveryFilesService):
    """
    Fetch the raw data files for a case from Housekeeper.
    """

    def __init__(
        self,
        status_db: Store,
        hk_api: HousekeeperAPI,
        tags_fetcher: SampleAndCaseDeliveryTagsFetcher,
    ):
        self.status_db = status_db
        self.hk_api = hk_api
        self.tags_fetcher = tags_fetcher

    def get_files_to_deliver(self, case_id: str, sample_id: str | None = None) -> DeliveryFiles:
        """
        Return a list of raw data files to be delivered for a case and its samples.
        args:
            case_id: The case id to deliver files for
            sample_id: The sample id to deliver files for
        """
        LOG.debug(f"[FETCH SERVICE] Fetching raw data files for case: {case_id}")
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_ids: list[str] = [sample_id] if sample_id else case.sample_ids
        raw_data_files: list[SampleFile] = []
        for sample_id in sample_ids:
            raw_data_files.extend(
                self._get_raw_data_files_for_sample(case_id=case_id, sample_id=sample_id)
            )
        delivery_data = DeliveryMetaData(
            case_id=case.internal_id,
            customer_internal_id=case.customer.internal_id,
            ticket_id=case.latest_ticket,
        )
        return self._validate_delivery_has_content(
            DeliveryFiles(
                delivery_data=delivery_data,
                case_files=[],
                sample_files=raw_data_files,
            )
        )

    @staticmethod
    def _validate_delivery_has_content(delivery_files: DeliveryFiles) -> DeliveryFiles:
        """Check if the delivery files has files to deliver.
        raises:
            NoDeliveryFilesError if no files to deliver.
        args:
            delivery_files: The delivery files to check
        """
        for sample_file in delivery_files.sample_files:
            LOG.debug(
                f"Found file to deliver: {sample_file.file_path} for sample: {sample_file.sample_id}"
            )
        if delivery_files.sample_files:
            return delivery_files
        LOG.info(
            f"No files to deliver for case {delivery_files.delivery_data.case_id} in ticket: {delivery_files.delivery_data.ticket_id}"
        )
        raise NoDeliveryFilesError

    @handle_missing_bundle_errors
    def _get_raw_data_files_for_sample(self, case_id: str, sample_id: str) -> list[SampleFile]:
        """
        Get the RawData files for a sample. Hardcoded tags to fetch from the raw data workflow.
        args:
            case_id: The case id to get the raw data files for
            sample_id: The sample id to get the raw data files for
        """
        file_tags: list[set[str]] = self.tags_fetcher.fetch_tags(Workflow.RAW_DATA).sample_tags
        raw_data_files: list[File] = self.hk_api.get_files_from_latest_version_containing_tags(
            bundle_name=sample_id, tags=file_tags
        )
        sample_name: str = self.status_db.get_sample_by_internal_id(sample_id).name
        return [
            SampleFile(
                case_id=case_id,
                sample_id=sample_id,
                sample_name=sample_name,
                file_path=raw_data_file.full_path,
            )
            for raw_data_file in raw_data_files
        ]



"""Taxprofiler delivery report API."""

from cg.constants import (
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_DATA_ANALYSIS_TAXPROFILER_FIELDS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_SAMPLE_METADATA_TAXPROFILER_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_TAXPROFILER_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
)
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.analysis import AnalysisModel, NextflowAnalysis
from cg.models.delivery_report.metadata import TaxprofilerSampleMetadataModel
from cg.models.delivery_report.report import CaseModel, ReportRequiredFields
from cg.models.delivery_report.sample import SampleModel
from cg.store.models import Case, Sample


class TaxprofilerDeliveryReportAPI(DeliveryReportAPI):
    """API to create Taxprofiler delivery reports."""

    def __init__(self, analysis_api: TaxprofilerAnalysisAPI):
        super().__init__(analysis_api=analysis_api)

    def get_sample_metadata(
        self, case: Case, sample: Sample, analysis_metadata: NextflowAnalysis
    ) -> TaxprofilerSampleMetadataModel:
        """Return Taxprofiler sample metadata to include in the delivery report."""
        return TaxprofilerSampleMetadataModel(
            initial_qc=self.lims_api.has_sample_passed_initial_qc(sample.internal_id)
        )

    def is_report_accredited(
        self, samples: list[SampleModel], analysis_metadata: AnalysisModel
    ) -> bool:
        """Return whether the Taxprofiler delivery report is accredited."""
        return False

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return the delivery report required fields for Taxprofiler."""

        report_required_fields = ReportRequiredFields(
            applications=self.get_application_required_fields(
                case=case, required_fields=REQUIRED_APPLICATION_FIELDS
            ),
            case=REQUIRED_CASE_FIELDS,
            customer=REQUIRED_CUSTOMER_FIELDS,
            data_analysis=REQUIRED_DATA_ANALYSIS_TAXPROFILER_FIELDS,
            metadata=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METADATA_TAXPROFILER_FIELDS
            ),
            methods=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METHODS_FIELDS
            ),
            report=REQUIRED_REPORT_FIELDS,
            samples=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TAXPROFILER_FIELDS
            ),
            timestamps=self.get_timestamp_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TIMESTAMP_FIELDS
            ),
        )
        return report_required_fields.model_dump()



from cg.apps.lims import LimsAPI
from cg.models.orders.constants import OrderType
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.implementations.case_order_service import StoreCaseOrderService
from cg.services.orders.storing.implementations.fastq_order_service import StoreFastqOrderService
from cg.services.orders.storing.implementations.metagenome_order_service import (
    StoreMetagenomeOrderService,
)
from cg.services.orders.storing.implementations.microbial_fastq_order_service import (
    StoreMicrobialFastqOrderService,
)
from cg.services.orders.storing.implementations.microbial_order_service import (
    StoreMicrobialOrderService,
)
from cg.services.orders.storing.implementations.pacbio_order_service import StorePacBioOrderService
from cg.services.orders.storing.implementations.pool_order_service import StorePoolOrderService
from cg.services.orders.storing.implementations.taxprofiler_order_service import (
    StoreTaxprofilerOrderService,
)
from cg.services.orders.storing.service import StoreOrderService
from cg.store.store import Store


class StoringServiceRegistry:
    """
    A registry for StoreOrderService instances, keyed by OrderType.
    """

    def __init__(self):
        self._registry = {}

    def register(self, order_type: OrderType, storing_service: StoreOrderService):
        """Register a StoreOrderService instance for a given OrderType."""
        self._registry[order_type] = storing_service

    def get_storing_service(self, order_type: OrderType) -> StoreOrderService:
        """Fetch the registered StoreOrderService for the given OrderType."""
        if storing_service := self._registry.get(order_type):
            return storing_service
        raise ValueError(f"No StoreOrderService registered for order type: {order_type}")


order_service_mapping = {
    OrderType.BALSAMIC: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.BALSAMIC_UMI: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.FASTQ: (
        OrderLimsService,
        StoreFastqOrderService,
    ),
    OrderType.FLUFFY: (
        OrderLimsService,
        StorePoolOrderService,
    ),
    OrderType.METAGENOME: (
        OrderLimsService,
        StoreMetagenomeOrderService,
    ),
    OrderType.MICROBIAL_FASTQ: (
        OrderLimsService,
        StoreMicrobialFastqOrderService,
    ),
    OrderType.MICROSALT: (
        OrderLimsService,
        StoreMicrobialOrderService,
    ),
    OrderType.MIP_DNA: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.MIP_RNA: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.NALLO: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.PACBIO_LONG_READ: (
        OrderLimsService,
        StorePacBioOrderService,
    ),
    OrderType.RAREDISEASE: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.RML: (
        OrderLimsService,
        StorePoolOrderService,
    ),
    OrderType.RNAFUSION: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
    OrderType.SARS_COV_2: (
        OrderLimsService,
        StoreMicrobialOrderService,
    ),
    OrderType.TAXPROFILER: (
        OrderLimsService,
        StoreTaxprofilerOrderService,
    ),
    OrderType.TOMTE: (
        OrderLimsService,
        StoreCaseOrderService,
    ),
}


def build_storing_service(
    lims: LimsAPI, status_db: Store, order_type: OrderType
) -> StoreOrderService:
    """Build a StoreOrderService instance for the given OrderType."""
    lims_service, store_service = order_service_mapping[order_type]
    return store_service(status_db, lims_service(lims))


def setup_storing_service_registry(lims: LimsAPI, status_db: Store) -> StoringServiceRegistry:
    """Set up the StoringServiceRegistry with all StoreOrderService instances."""
    registry = StoringServiceRegistry()
    for order_type in order_service_mapping.keys():
        registry.register(
            order_type=order_type,
            storing_service=build_storing_service(
                lims=lims, status_db=status_db, order_type=order_type
            ),
        )
    return registry



from pydantic import BeforeValidator, Field
from typing_extensions import Annotated

from cg.models.orders.sample_base import NAME_PATTERN, PriorityEnum, SexEnum
from cg.services.orders.validation.constants import ElutionBuffer
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.utils import parse_buffer


class FastqSample(Sample):
    capture_kit: str | None = None
    concentration_ng_ul: float | None = None
    elution_buffer: Annotated[ElutionBuffer, BeforeValidator(parse_buffer)]
    priority: PriorityEnum
    quantity: int | None = None
    require_qc_ok: bool = False
    sex: SexEnum
    source: str
    subject_id: str = Field(pattern=NAME_PATTERN, max_length=128)
    tumour: bool = False



from flask import Blueprint, jsonify, request

from cg.server.endpoints.sequencing_metrics.dtos import (
    PacbioSequencingMetricsRequest,
    PacbioSequencingMetricsResponse,
)
from cg.server.endpoints.sequencing_metrics.error_handler import handle_endpoint_errors
from cg.server.endpoints.utils import before_request
from cg.server.ext import sample_run_metrics_service
from cg.services.sample_run_metrics_service.dtos import PacbioSequencingMetricsDTO

PACBIO_SAMPLE_SEQUENCING_METRICS_BLUEPRINT = Blueprint(
    "pacbio_sample_sequencing_metrics", __name__, url_prefix="/api/v1"
)
PACBIO_SAMPLE_SEQUENCING_METRICS_BLUEPRINT.before_request(before_request)


@PACBIO_SAMPLE_SEQUENCING_METRICS_BLUEPRINT.route("/pacbio_sample_sequencing_metrics")
@handle_endpoint_errors
def get_sequencing_metrics():
    sequencing_metrics_request = PacbioSequencingMetricsRequest(
        sample_id=request.args.get("sample_id"), smrt_cell_ids=request.args.getlist("smrt_cell_ids")
    )
    metrics: list[PacbioSequencingMetricsDTO] = sample_run_metrics_service.get_pacbio_metrics(
        sequencing_metrics_request
    )
    response = PacbioSequencingMetricsResponse(metrics=metrics)
    return jsonify(response.model_dump())



from pathlib import Path
from cg.apps.lims.api import LimsAPI
from cg.constants.constants import MutantQC
from cg.constants.lims import LimsProcess
from cg.exc import CgError
from cg.meta.workflow.mutant.quality_controller.metrics_parser_utils import parse_samples_results
from cg.meta.workflow.mutant.quality_controller.models import (
    MutantPoolSamples,
    SamplePoolAndResults,
    SampleQualityResults,
    CaseQualityResult,
    MutantQualityResult,
    ParsedSampleResults,
    SamplesQualityResults,
)
from cg.meta.workflow.mutant.quality_controller.report_generator_utils import (
    get_summary,
    write_report,
)
from cg.meta.workflow.mutant.quality_controller.result_logger_utils import (
    log_case_result,
    log_results,
    log_sample_result,
)
from cg.meta.workflow.mutant.quality_controller.utils import (
    has_external_negative_control_sample_valid_total_reads,
    has_internal_negative_control_sample_valid_total_reads,
    has_sample_valid_total_reads,
)
from cg.store.models import Case, Sample
from cg.store.store import Store


class MutantQualityController:
    def __init__(self, status_db: Store, lims: LimsAPI) -> None:
        self.status_db: Store = status_db
        self.lims: LimsAPI = lims

    def get_quality_control_result(
        self, case: Case, case_results_file_path: Path, case_qc_report_path: Path
    ) -> MutantQualityResult:
        """Perform QC check on a case and generate the QC_report."""
        sample_pool_and_results: SamplePoolAndResults = self._get_sample_pool_and_results(
            case_results_file_path=case_results_file_path,
            case=case,
        )

        samples_quality_results: SamplesQualityResults = self._get_samples_quality_results(
            sample_pool_and_results=sample_pool_and_results
        )
        case_quality_result: CaseQualityResult = self._get_case_quality_result(
            samples_quality_results
        )

        write_report(
            case_qc_report_path=case_qc_report_path,
            samples_quality_results=samples_quality_results,
            case_quality_result=case_quality_result,
        )

        log_results(
            case_quality_result=case_quality_result,
            samples_quality_results=samples_quality_results,
            report_file_path=case_qc_report_path,
        )

        summary: str = get_summary(
            case_quality_result=case_quality_result,
            samples_quality_results=samples_quality_results,
        )

        return MutantQualityResult(
            case_quality_result=case_quality_result,
            samples_quality_results=samples_quality_results,
            summary=summary,
        )

    def _get_samples_quality_results(
        self, sample_pool_and_results: SamplePoolAndResults
    ) -> SamplesQualityResults:
        samples_quality_results: list[SampleQualityResults] = []
        for sample in sample_pool_and_results.pool.samples:
            sample_results: ParsedSampleResults = sample_pool_and_results.results[
                sample.internal_id
            ]
            sample_quality_results: SampleQualityResults = (
                self._get_sample_quality_result_for_sample(
                    sample=sample, sample_results=sample_results
                )
            )
            samples_quality_results.append(sample_quality_results)

        internal_negative_control_sample: Sample = (
            sample_pool_and_results.pool.internal_negative_control
        )
        internal_negative_control_quality_metrics: SampleQualityResults = (
            self._get_sample_quality_result_for_internal_negative_control_sample(
                sample=internal_negative_control_sample
            )
        )

        external_negative_control_sample: Sample = (
            sample_pool_and_results.pool.external_negative_control
        )
        external_negative_control_sample_results: ParsedSampleResults = (
            sample_pool_and_results.results[external_negative_control_sample.internal_id]
        )
        external_negative_control_quality_metrics: SampleQualityResults = (
            self._get_sample_quality_result_for_external_negative_control_sample(
                sample=external_negative_control_sample,
                sample_results=external_negative_control_sample_results,
            )
        )

        return SamplesQualityResults(
            samples=samples_quality_results,
            internal_negative_control=internal_negative_control_quality_metrics,
            external_negative_control=external_negative_control_quality_metrics,
        )

    @staticmethod
    def _get_sample_quality_result_for_sample(
        sample: Sample, sample_results: ParsedSampleResults
    ) -> SampleQualityResults:
        does_sample_pass_reads_threshold: bool = has_sample_valid_total_reads(sample=sample)
        does_sample_pass_qc: bool = does_sample_pass_reads_threshold and sample_results.passes_qc
        sample_quality_result = SampleQualityResults(
            sample_id=sample.internal_id,
            passes_qc=does_sample_pass_qc,
            passes_reads_threshold=does_sample_pass_reads_threshold,
            passes_mutant_qc=sample_results.passes_qc,
        )

        log_sample_result(
            result=sample_quality_result,
        )
        return sample_quality_result

    @staticmethod
    def _get_sample_quality_result_for_internal_negative_control_sample(
        sample: Sample,
    ) -> SampleQualityResults:
        does_sample_pass_reads_threshold: bool = (
            has_internal_negative_control_sample_valid_total_reads(sample=sample)
        )
        sample_quality_result = SampleQualityResults(
            sample_id=sample.internal_id,
            passes_qc=does_sample_pass_reads_threshold,
            passes_reads_threshold=does_sample_pass_reads_threshold,
        )

        log_sample_result(result=sample_quality_result, is_external_negative_control=True)
        return sample_quality_result

    @staticmethod
    def _get_sample_quality_result_for_external_negative_control_sample(
        sample: Sample, sample_results: ParsedSampleResults
    ) -> SampleQualityResults:
        does_sample_pass_reads_threshold: bool = (
            has_external_negative_control_sample_valid_total_reads(sample=sample)
        )
        sample_passes_qc: bool = does_sample_pass_reads_threshold and not sample_results.passes_qc
        sample_quality_result = SampleQualityResults(
            sample_id=sample.internal_id,
            passes_qc=sample_passes_qc,
            passes_reads_threshold=does_sample_pass_reads_threshold,
            passes_mutant_qc=sample_results.passes_qc,
        )

        log_sample_result(result=sample_quality_result, is_external_negative_control=True)
        return sample_quality_result

    def _get_case_quality_result(
        self, samples_quality_results: SamplesQualityResults
    ) -> CaseQualityResult:
        external_negative_control_pass_qc: bool = (
            samples_quality_results.external_negative_control.passes_qc
        )
        internal_negative_control_pass_qc: bool = (
            samples_quality_results.internal_negative_control.passes_qc
        )

        samples_pass_qc: bool = self._samples_pass_qc(
            samples_quality_results=samples_quality_results
        )

        case_passes_qc: bool = (
            samples_pass_qc
            and internal_negative_control_pass_qc
            and external_negative_control_pass_qc
        )

        result = CaseQualityResult(
            passes_qc=case_passes_qc,
            internal_negative_control_passes_qc=internal_negative_control_pass_qc,
            external_negative_control_passes_qc=external_negative_control_pass_qc,
            fraction_samples_passes_qc=samples_pass_qc,
        )

        log_case_result(result)
        return result

    @staticmethod
    def _samples_pass_qc(samples_quality_results: SamplesQualityResults) -> bool:
        fraction_failed_samples: float = (
            samples_quality_results.failed_samples_count
            / samples_quality_results.total_samples_count
        )
        return fraction_failed_samples < MutantQC.FRACTION_OF_SAMPLES_WITH_FAILED_QC_THRESHOLD

    def _get_internal_negative_control_id_for_case(self, case: Case) -> str:
        """Query lims to retrive internal_negative_control_id for a mutant case sequenced in one pool."""

        sample_internal_id = case.sample_ids[0]
        internal_negative_control_id: str = (
            self.lims.get_internal_negative_control_id_from_sample_in_pool(
                sample_internal_id=sample_internal_id, pooling_step=LimsProcess.COVID_POOLING_STEP
            )
        )
        return internal_negative_control_id

    def _get_internal_negative_control_sample_for_case(
        self,
        case: Case,
    ) -> Sample:
        internal_negative_control_id: str = self._get_internal_negative_control_id_for_case(
            case=case
        )
        return self.status_db.get_sample_by_internal_id(internal_id=internal_negative_control_id)

    def _get_mutant_pool_samples(self, case: Case) -> MutantPoolSamples:
        samples = []
        external_negative_control = None

        for sample in case.samples:
            if sample.is_negative_control:
                external_negative_control = sample
                continue
            samples.append(sample)

        if not external_negative_control:
            raise CgError(f"No external negative control sample found for case {case.internal_id}.")

        internal_negative_control: Sample = self._get_internal_negative_control_sample_for_case(
            case=case
        )

        return MutantPoolSamples(
            samples=samples,
            external_negative_control=external_negative_control,
            internal_negative_control=internal_negative_control,
        )

    def _get_sample_pool_and_results(
        self, case_results_file_path: Path, case: Case
    ) -> SamplePoolAndResults:
        try:
            samples: MutantPoolSamples = self._get_mutant_pool_samples(case=case)
        except Exception as exception_object:
            raise CgError(
                f"Not possible to retrieve samples for case {case.internal_id}: {exception_object}"
            ) from exception_object

        try:
            samples_results: dict[str, ParsedSampleResults] = parse_samples_results(
                case=case, results_file_path=case_results_file_path
            )
        except Exception as exception_object:
            raise CgError(
                f"Not possible to retrieve results for case {case.internal_id}: {exception_object}"
            )

        return SamplePoolAndResults(pool=samples, results=samples_results)



from pydantic import BaseModel, Field

from cg.models.orders.constants import OrderType


class CollaboratorSamplesRequest(BaseModel):
    customer: str
    enquiry: str
    limit: int = 50
    order_type: OrderType | None = None


class SamplesRequest(BaseModel):
    status: str | None = None
    enquiry: str | None = None
    page: int = 1
    page_size: int | None = Field(50, alias="pageSize")



from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.models.sample import Sample


class OrderWithSamples(Order):
    samples: list[Sample]

    @property
    def enumerated_samples(self) -> enumerate[Sample]:
        return enumerate(self.samples)



"""Loqusdb input files models."""

import logging

from pydantic import BaseModel, FilePath

LOG = logging.getLogger(__name__)


class ObservationsInputFiles(BaseModel):
    """Model for validating Loqusdb input files."""

    snv_vcf_path: FilePath
    sv_vcf_path: FilePath | None = None


class MipDNAObservationsInputFiles(ObservationsInputFiles):
    """Model for validating rare disease Loqusdb input files."""

    profile_vcf_path: FilePath
    family_ped_path: FilePath


class BalsamicObservationsInputFiles(ObservationsInputFiles):
    """Model for validating cancer Loqusdb input files."""

    snv_germline_vcf_path: FilePath
    sv_germline_vcf_path: FilePath | None = None


class NalloObservationsInputFiles(ObservationsInputFiles):
    """Model for validating Nallo Loqusdb input files."""

    profile_vcf_path: FilePath
    family_ped_path: FilePath


class RarediseaseObservationsInputFiles(ObservationsInputFiles):
    """Model for validating RAREDISEASE Loqusdb input files."""

    profile_vcf_path: FilePath
    family_ped_path: FilePath



from enum import Enum


class DeliveryDestination(Enum):
    """Enum for the DeliveryDestination
    BASE: Deliver to the base folder provided in the call
    CUSTOMER: Deliver to the customer folder on hasta
    FOHM: Deliver to the FOHM folder on hasta
    """

    BASE = "base"
    CUSTOMER = "customer"
    FOHM = "fohm"


class DeliveryStructure(Enum):
    """Enum for the DeliveryStructure
    FLAT: Deliver the files in a flat structure, i.e. all files in the same folder
    NESTED: Deliver the files in a nested structure, i.e. files in folders for each sample/case
    """

    FLAT: str = "flat"
    NESTED: str = "nested"



from pydantic import BaseModel
from typing_extensions import Literal


class SbatchError(BaseModel):
    flow_cell_id: str
    email: str
    logfile: str
    demux_dir: str
    demux_started: str


class SbatchCommand(BaseModel):
    run_dir: str  # path/to/a_flowcell/
    demux_dir: str  # path/to/output_dir/
    unaligned_dir: str | None = None  # path/to/output_dir/Unaligned/
    sample_sheet: str  # path/to/SampleSheet.csv
    demux_completed_file: str  # path/to/demuxcomplete.txt
    environment: Literal["stage", "production"]



from enum import StrEnum, auto


class TissueBlockEnum(StrEnum):
    SMALL: str = auto()
    LARGE: str = auto()
    BLANK: str = ""


class ElutionBuffer(StrEnum):
    """The choices of buffers."""

    OTHER = "Other"
    TRIS_HCL = "Tris-HCl"
    WATER = "Nuclease-free water"


ALLOWED_SKIP_RC_BUFFERS = [ElutionBuffer.TRIS_HCL, ElutionBuffer.WATER]

MINIMUM_VOLUME, MAXIMUM_VOLUME = 20, 130


class ExtractionMethod(StrEnum):
    EZ1 = "EZ1"
    MAELSTROM = "Maelstrom"
    MAGNAPURE_96 = "MagNaPure 96"
    QIAGEN_MAGATTRACT = "Qiagen MagAttract"
    QIASYMPHONE = "QIAsymphony"
    OTHER = 'Other (specify in "Comments")'


class IndexEnum(StrEnum):
    ARCHER_SET_B = "Archer Set B"
    AVIDA_INDEX_PLATE = "Avida Index plate"
    AVIDA_INDEX_STRIP = "Avida Index strip"
    IDT_DS_B = "IDT DupSeq 10 bp Set B"
    IDT_DS_F = "IDT DupSeq 10 bp Set F"
    IDT_XGEN_UDI = "IDT xGen UDI Adapters"
    IDT10_UDI_NIPT = "IDT10 UDI NIPT"
    KAPA_UDI_NIPT = "KAPA UDI NIPT"
    NEBNEXT_LV_Unique_Dual_2A = "NEBNext LV Unique Dual Index Primers Set 2A"
    NEXTERA_XT = "Nextera XT Dual"
    NEXTFLEX_UDI_96 = "NEXTflex® Unique Dual Index Barcodes 1 - 96"
    NEXTFLEX_V2_UDI_96 = "NEXTflex® v2 UDI Barcodes 1 - 96"
    TEN_X_TN_A = "10X Genomics Dual Index kit TN Set A"
    TEN_X_TT_A = "10X Genomics Dual Index kit TT Set A"
    TWIST_UDI_A = "TWIST UDI Set A"
    TWIST_UDI_B = "TWIST UDI Set B"
    TWIST_UDI_C = "TWIST UDI Set C"
    TRUSEQ_DNA_HT = "TruSeq DNA HT Dual-index (D7-D5)"
    XGEN_UDI_10_1_4 = "xGen UDI 10nt Primer Plates 1-4"
    NO_INDEX = "NoIndex"



from cg.constants.constants import SexOptions
from cg.models.qc_metrics import QCMetrics


class RarediseaseQCMetrics(QCMetrics):
    """Raredisease QC metrics."""

    mapped_reads: int
    percent_duplication: float
    predicted_sex_sex_check: SexOptions
    total_reads: int



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.balsamic.models.case import BalsamicCase
from cg.services.orders.validation.order_types.balsamic_umi.models.sample import BalsamicUmiSample

NewSample = Annotated[BalsamicUmiSample, Tag("new")]
OldSample = Annotated[ExistingSample, Tag("existing")]


class BalsamicUmiCase(BalsamicCase):
    samples: list[Annotated[NewSample | OldSample, Discriminator(has_internal_id)]]



from pathlib import Path

from cg.constants import FileExtensions
from cg.utils.files import get_project_root_dir

project_root_dir: Path = get_project_root_dir()

NALLO_BUNDLE_FILENAMES: str = (
    Path("resources", "nallo_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)
NALLO_BUNDLE_FILENAMES_PATH = Path(project_root_dir, NALLO_BUNDLE_FILENAMES)

RAREDISEASE_BUNDLE_FILENAMES: str = (
    Path("resources", "raredisease_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)
RAREDISEASE_BUNDLE_FILENAMES_PATH = Path(project_root_dir, RAREDISEASE_BUNDLE_FILENAMES)

RNAFUSION_BUNDLE_FILENAMES: str = (
    Path("resources", "rnafusion_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)
RNAFUSION_BUNDLE_FILENAMES_PATH = Path(project_root_dir, RNAFUSION_BUNDLE_FILENAMES)

TAXPROFILER_BUNDLE_FILENAMES: str = (
    Path("resources", "taxprofiler_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)
TAXPROFILER_BUNDLE_FILENAMES_PATH = Path(project_root_dir, TAXPROFILER_BUNDLE_FILENAMES)

TOMTE_BUNDLE_FILENAMES: str = (
    Path("resources", "tomte_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)
TOMTE_BUNDLE_FILENAMES_PATH = Path(project_root_dir, TOMTE_BUNDLE_FILENAMES)



"""Module for Jasen Analysis API."""

import logging

from cg.constants import Workflow
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class JasenAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Jasen processes and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.JASEN,
    ):
        super().__init__(config=config, workflow=workflow)



from pathlib import Path

from pydantic import BaseModel, field_validator


class NfCommandArgs(BaseModel):
    """Model for arguments and options supported."""

    log: str | Path | None = None
    resume: bool | None = None
    profile: str | None = None
    stub_run: bool | None = None
    config: str | Path | None = None
    name: str | None = None
    revision: str | None = None
    wait: str | None = None
    id: str | None = None
    with_tower: bool | None = None
    use_nextflow: bool | None = None
    compute_env: str | None = None
    work_dir: str | Path | None = None
    params_file: str | Path | None = None


class FileDeliverable(BaseModel):
    """Specification for a general deliverables file."""

    id: str
    format: str
    path: str
    path_index: str | None = None
    step: str
    tag: str

    @field_validator("path", "path_index", mode="before")
    @classmethod
    def set_path_as_string(cls, file_path: str | Path) -> str | None:
        return str(Path(file_path)) if file_path else None


class WorkflowDeliverables(BaseModel):
    """Specification for workflow deliverables."""

    files: list[FileDeliverable]



from pydantic import BaseModel

from cg.services.orders.validation.errors.case_errors import CaseError
from cg.services.orders.validation.errors.case_sample_errors import CaseSampleError
from cg.services.orders.validation.errors.order_errors import OrderError
from cg.services.orders.validation.errors.sample_errors import SampleError


class ValidationErrors(BaseModel):
    order_errors: list[OrderError] = []
    case_errors: list[CaseError] = []
    sample_errors: list[SampleError] = []
    case_sample_errors: list[CaseSampleError] = []

    @property
    def is_empty(self) -> bool:
        """Return True if there are no errors in any of the attributes,
        or if all the errors are actually warnings."""
        return all(
            not getattr(self, field)
            or (all(getattr(item, "field") == "warnings" for item in getattr(self, field)))
            for field in self.model_fields
        )

    def get_error_message(self) -> str:
        """Gets a string documenting all errors."""
        error_string = ""
        for error in self.order_errors:
            error_string += f"Problem with {error.field}: {error.message} \n"
        for error in self.case_errors:
            error_string += (
                f"Problem with {error.field} in case {error.case_index}: {error.message} \n"
            )
        for error in self.case_sample_errors:
            error_string += f"Problem with {error.field} in case {error.case_index} sample {error.sample_index}: {error.message} \n"
        for error in self.sample_errors:
            error_string += (
                f"Problem with {error.field} in sample {error.sample_index}: {error.message} \n"
            )
        return error_string



import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterator

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.io.csv import write_csv
from cg.io.gzip import read_gzip_first_line
from cg.meta.workflow.fastq import is_undetermined_in_path
from cg.models.fastq import FastqFileMeta, GetFastqFileMeta
from cg.store.models import Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class NextflowFastqSampleSheetCreator(ABC):

    def __init__(self, housekeeper_api: HousekeeperAPI, store: Store):
        self.housekeeper_api = housekeeper_api
        self.store = store

    def create(self, case_id: str, file_path: Path) -> None:
        LOG.debug(f"Creating sample sheet for case {case_id}")
        content: list[list[str]] = self._get_content(case_id)
        write_csv(file_path=file_path, content=content)

    @abstractmethod
    def _get_content(self, case_id: str) -> list[list[str]]:
        pass

    def _get_paired_read_paths(self, sample: Sample) -> Iterator[tuple[str, str]]:
        """Return an iterator of tuples with paired fastq paths for the forward and reverse read."""
        sample_metadata: list[FastqFileMeta] = self._get_fastq_metadata_for_sample(sample)
        fastq_forward_read_paths: list[str] = self._extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self._extract_read_files(
            metadata=sample_metadata, forward_read=False
        )
        return zip(fastq_forward_read_paths, fastq_reverse_read_paths)

    def _get_fastq_metadata_for_sample(self, sample: Sample) -> list[FastqFileMeta]:
        """Return FASTQ metadata objects for all fastq files linked to a sample."""
        return [
            self._parse_fastq_data(hk_file.full_path)
            for hk_file in self.housekeeper_api.files(
                bundle=sample.internal_id, tags={SequencingFileTag.FASTQ}
            )
        ]

    def _parse_fastq_data(self, fastq_path: Path) -> FastqFileMeta:
        header_line: str = read_gzip_first_line(file_path=fastq_path)
        fastq_file_meta: FastqFileMeta = self._parse_fastq_header(header_line)
        fastq_file_meta.path = fastq_path
        fastq_file_meta.undetermined = is_undetermined_in_path(fastq_path)
        matches = re.findall(r"-l[1-9]t([1-9]{2})_", str(fastq_path))
        if len(matches) > 0:
            fastq_file_meta.flow_cell_id = f"{fastq_file_meta.flow_cell_id}-{matches[0]}"
        return fastq_file_meta

    @staticmethod
    def _parse_fastq_header(line: str) -> FastqFileMeta | None:
        """Parse and return fastq header metadata.
        Handle Illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format
        Raise:
            TypeError if unable to split line into expected parts.
        """
        parts = line.split(":")
        try:
            return GetFastqFileMeta.header_format.get(len(parts))(parts=parts)
        except TypeError as exception:
            LOG.error(f"Could not parse header format for header: {line}")
            raise exception

    @staticmethod
    def _extract_read_files(metadata: list[FastqFileMeta], forward_read: bool) -> list[str]:
        """Extract a list of fastq file paths for either forward or reverse reads."""
        read_direction: int = 1 if forward_read else 2
        sorted_metadata: list = sorted(metadata, key=lambda k: k.path)
        return [
            fastq_file.path
            for fastq_file in sorted_metadata
            if fastq_file.read_direction == read_direction
        ]



from pydantic import BaseModel

from cg.constants import Workflow


class CaseConfig(BaseModel):
    case_id: str
    workflow: Workflow

    def get_session_id(self) -> str | None:
        # This will be set by the NextflowCaseConfig
        return None

    def get_workflow_id(self) -> str | None:
        # This will be set by the NextflowCaseConfig
        return None
